#!/usr/bin/env python3

# Copyright © 2022 erzo <erzo@posteo.de>
# This work is free. You can use, copy, modify, and/or distribute it
# under the terms of the BSD Zero Clause License, see LICENSE.

"""
usage: gits [<options>] [<commit>]

This is a wrapper around `git show` using the details view from gitl.
It takes the following command line arguments:

{doc_common_options}

All other command line arguments are passed through to git show.
You can use that for example in your vimrc:

    map <c-c> :!gits <C-r><C-w><cr><cr>

With that line pressing control + c opens the commit whose hash id is under the cursor.
This is useful while writing a commit message if you want to verify that you have copied
the correct hash or when doing an interactive rebase.

For more information see `git show --help`.
"""

from . import main
from . import check

main.DetailsView.max_lines_per_file = 2000


def run():
	import sys

	args = sys.argv[1:]
	kw = main.parse_common_args(args, doc=__doc__)

	if not args:
		diff_args = []
		hash_id = "HEAD"
	else:
		diff_args = args[:-1]
		hash_id = args[-1]

	a = main.App(show=hash_id, diff_args=diff_args, **kw)

	while a.continue_running:
		a.run()
		a.run_external_if_requested()


if __name__ == '__main__':
	run()



#!/usr/bin/env python3

import urwid

from . import model_input_history

class EditWithHistory(urwid.Edit):

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.history = model_input_history.History()
		self.command_map = self._command_map.copy()
		self._command_map = self._command_map.copy()
		self._command_map.clear_command(urwid.CURSOR_UP)
		self._command_map.clear_command(urwid.CURSOR_DOWN)

	def keypress(self, size, key):
		if not super().keypress(size, key):
			return None

		cmd = self.command_map[key]

		if cmd == urwid.CURSOR_UP:
			if self.history.stash_and_prev(self.edit_text, self.edit_pos):
				self.edit_text = self.history.get_text()
				self.edit_pos = self.history.get_cursor_position()
			return None

		elif cmd == urwid.CURSOR_DOWN:
			if self.history.stash_and_next(self.edit_text, self.edit_pos):
				self.edit_text = self.history.get_text()
				self.edit_pos = self.history.get_cursor_position()
			return None

		if cmd == urwid.ACTIVATE:
			self.history.commit(self.edit_text, self.edit_pos)
			return key

		return key

	def get_last_input(self):
		return self.history.get_text()


if __name__ == '__main__':
	def keypress(key):
		if key == 'esc':
			raise urwid.ExitMainLoop()
		elif urwid.command_map[key] == urwid.ACTIVATE:
			edit.edit_text = ""
		else:
			statusbar.set_text("unhandled key: %s" % key)

	edit = EditWithHistory()
	widget = urwid.Filler(edit)
	statusbar = urwid.Text("")
	frame = urwid.Frame(widget, footer=statusbar)
	urwid.MainLoop(frame, unhandled_input=keypress).run()



#!/usr/bin/env python3

from . import urwid_constants
from . import color_decoder

CD = color_decoder.ColorDecoder
COLOR = urwid_constants.COLOR

default_foreground_color = {
	CD.FG_DEFAULT: COLOR.FG_DEFAULT,
}

standard_foreground_colors = {
	CD.FG_BLACK:   COLOR.FG_BLACK,
	CD.FG_RED:     COLOR.FG_RED,
	CD.FG_GREEN:   COLOR.FG_GREEN,
	CD.FG_YELLOW:  COLOR.FG_YELLOW,
	CD.FG_BLUE:    COLOR.FG_BLUE,
	CD.FG_MAGENTA: COLOR.FG_MAGENTA,
	CD.FG_CYAN:    COLOR.FG_CYAN,
	CD.FG_GRAY:    COLOR.FG_WHITE,
}

light_foreground_colors = {
	CD.FG_LIGHT_BLACK:   COLOR.FG_BRIGHT_BLACK,
	CD.FG_LIGHT_RED:     COLOR.FG_BRIGHT_RED,
	CD.FG_LIGHT_GREEN:   COLOR.FG_BRIGHT_GREEN,
	CD.FG_LIGHT_YELLOW:  COLOR.FG_BRIGHT_YELLOW,
	CD.FG_LIGHT_BLUE:    COLOR.FG_BRIGHT_BLUE,
	CD.FG_LIGHT_MAGENTA: COLOR.FG_BRIGHT_MAGENTA,
	CD.FG_LIGHT_CYAN:    COLOR.FG_BRIGHT_CYAN,
	CD.FG_LIGHT_GRAY:    COLOR.FG_BRIGHT_WHITE,
}

default_background_color = {
	CD.BG_DEFAULT: COLOR.BG_DEFAULT,
}

background_colors = {
	CD.BG_BLACK:   COLOR.BG_BLACK,
	CD.BG_RED:     COLOR.BG_RED,
	CD.BG_GREEN:   COLOR.BG_GREEN,
	CD.BG_YELLOW:  COLOR.BG_YELLOW,
	CD.BG_BLUE:    COLOR.BG_BLUE,
	CD.BG_MAGENTA: COLOR.BG_MAGENTA,
	CD.BG_CYAN:    COLOR.BG_CYAN,
	CD.BG_GRAY:    COLOR.BG_WHITE,
}

series = {
	CD.NORMAL : "",
	CD.BOLD   : urwid_constants.EMPH.BOLD,
}


class Generator:

	"""
	Create a pallete which maps the color_decoder names to urwid colors.

	For efficiency reasons not all possible combinations are created
	by default. You can use the arguments of the constructor to select
	which colors/formats you need.

	If you set fg to True and bg to True all possible combinations
	between the two will be created.
	If you want to have foreground colors on the default background
	and the default foreground on background colors but not all
	possible combinations of foreground and background colors
	use two Generator objects (where one has fg=True, bg=False and
	one has fg=False, bg=True) and add their output.
	"""

	FOCUS_SUFFIX = "_focus"

	def __init__(self, bold=True, standard_fg=True, light_fg=False, bg=False, focus=True):
		self.activated_series = [CD.NORMAL]
		if bold:
			self.activated_series.append(CD.BOLD)

		self.activated_fg_colors = default_foreground_color
		if standard_fg:
			self.activated_fg_colors.update(standard_foreground_colors)
		if light_fg:
			self.activated_fg_colors.update(light_foreground_colors)

		self.activated_bg_colors = default_background_color
		if bg:
			self.activated_bg_colors.update(background_colors)

		self.focus = focus


	def palette(self):
		out = list()

		for series_decoder_name in self.activated_series:
			for bg_decoder_name, bg_urwid_name in self.activated_bg_colors.items():
				for fg_decoder_name, fg_urwid_name in self.activated_fg_colors.items():
					name = CD.combine_colors(series_decoder_name, fg_decoder_name, bg_decoder_name)
					if series_decoder_name == CD.BOLD:
						fg_urwid_name += urwid_constants.EMPH.BOLD
					out.append((name, fg_urwid_name, bg_urwid_name))

					if self.focus:
						name += self.FOCUS_SUFFIX
						fg_urwid_name += urwid_constants.EMPH.STANDOUT
						out.append((name, fg_urwid_name, bg_urwid_name))

		return out


	def focus_map(self):
		out = dict()

		for series_decoder_name in self.activated_series:
			for bg_decoder_name, bg_urwid_name in self.activated_bg_colors.items():
				for fg_decoder_name, fg_urwid_name in self.activated_fg_colors.items():
					name = CD.combine_colors(series_decoder_name, fg_decoder_name, bg_decoder_name)
					out[name] = name + self.FOCUS_SUFFIX

		return out


if __name__ == '__main__':
	g = Generator()
	palette = g.palette()

	import sys
	lines = []
	for fn in sys.argv[1:]:
		with open(fn, 'rt') as f:
			lines.extend(f.readlines())

	if lines:
		# test usage:
		# $ python urwid_color_encoder.py <(colortest)
		# should look pretty much the same like
		# $ colortest

		import urwid

		def quit(key):
			if key.lower() == 'q':
				raise urwid.ExitMainLoop()

		focus_map = g.focus_map()
		decoder = color_decoder.ColorDecoder()
		view = urwid.ListBox([urwid.AttrMap(urwid.SelectableIcon(decoder.decode(ln.rstrip()) + ['']), None, focus_map=focus_map) for ln in lines])
		urwid.MainLoop(view, palette=palette, unhandled_input=quit).run()
	
	else:
		print("\n".join("%-30s, %-30s, %-30s" %(i) for i in palette))



#!/usr/bin/env python3

import urwid
calc_width = urwid.util.calc_width
calc_text_pos = urwid.util.calc_text_pos
move_next_char = urwid.util.move_next_char
move_prev_char = urwid.util.move_prev_char


def reverse_calc_text_pos(text, i0, i1, max_width):
	'''
	return (i, sc)
	where i is the smallest text index with i0 <= i <= i1
	so that text[i:i1] takes up as many screen column sc
	as possible with sc <= max_width. (If text contains
	full width characters sc may be smaller than max_width.)
	'''
	sc = 0
	while sc < max_width and i1 > i0:
		i = move_prev_char(text, i0, i1)
		char_width = calc_width(text, i, i1)
		if sc + char_width > max_width:
			break
		i1 = i
		sc += char_width
	return i1, sc


class LogTextLayout(urwid.StandardTextLayout):

	"""
	Extends the StandardTextLayout by a new align mode:
	ALIGN_INDENTATION

	In this mode text will be printed left aligned
	with the first line starting at the very left
	and the following lines indented.
	The indentation is specified by the first occurence
	of align_character in each printed line. This first
	occurence of align_character is hidden.
	Spaces directly following the align_character become
	part of the indentation as well because the %d place
	holder of git log expands to either nothing or some
	thing starting with " (" so that it is not possible
	to put the align_character after the space.

	Also extends the StandardTextLayout by a new wrap mode:
	WRAP_CUSTOM

	This is a generalization of WRAP_SPACE. The characters
	after which a break is allowed can be specified with
	wrap_after_characters.
	"""

	ALIGN_LEFT = urwid.LEFT
	ALIGN_RIGHT = urwid.RIGHT
	ALIGN_CENTER = urwid.CENTER
	ALIGN_INDENTATION = "left indentation"
	ALLOWED_ALIGN_VALUES = (ALIGN_LEFT, ALIGN_RIGHT, ALIGN_CENTER, ALIGN_INDENTATION)

	WRAP_ANY   = urwid.ANY
	WRAP_SPACE = urwid.SPACE
	WRAP_CLIP  = urwid.CLIP
	WRAP_ELLIPSIS = 'ellipsis'
	WRAP_CUSTOM   = 'custom'
	ALLOWED_WRAP_VALUES = (WRAP_ANY, WRAP_SPACE, WRAP_CLIP, WRAP_ELLIPSIS, WRAP_CUSTOM)

	# minimum number of screen columns for displaying text (with ALIGN_INDENTATION)
	# If the window is too small to show indentation + MIN_WIDTH the indentation is omitted.
	# This value must be greater than 1. Otherwise full width characters and everything after them
	# will not be display if the window has a width of indentation + 1.
	MIN_WIDTH = 2

	EXPLICIT_SPACES = False

	def __init__(self, align_character, ellipsis_character="…".encode('utf-8'), wrap_after_characters="- ", wrap_before_characters=" ", repeat_prefix=False, auto_move_align_character=True):
		"""
		- align_character [for align=ALIGN_INDENTATION]:
		  a string marking how for to indent the following lines
		- ellipsis_character [for wrap=WRAP_ELLIPSIS]:
		  a string to be appended to lines which are clipped because they are too long
		- wrap_after_characters [for wrap=WRAP_CUSTOM]:
		  lines can be wrapped after any character in this string
		- wrap_before_characters [for wrap=WRAP_CUSTOM]:
		  lines can be wrapped before any character in this string
		- repeat_prefix [for align=ALIGN_INDENTATION]:
		  False: broken lines are indented by inserting spaces
		  True: broken lines are indented by inserting the beginning of the text before align_character
		- auto_move_align_character [for align=ALIGN_INDENTATION]:
		  True: treat all space characters directly following the align_character as being before the align_character
		  False: take the position of the align_character literally
		"""
		super().__init__()
		self.align_character = align_character
		if isinstance(ellipsis_character, str):
			self.ellipsis_character = urwid.compat.B(ellipsis_character)
		else:
			self.ellipsis_character = ellipsis_character
		self.width_ellipsis_character = calc_width(ellipsis_character, 0, len(ellipsis_character))
		self.wrap_after_characters = wrap_after_characters
		self.wrap_before_characters = wrap_before_characters
		self.repeat_prefix = repeat_prefix
		self.auto_move_align_character = auto_move_align_character

	def copy(self, **kw):
		for arg in ('align_character', 'ellipsis_character', 'wrap_after_characters', 'wrap_before_characters', 'repeat_prefix', 'auto_move_align_character'):
			kw.setdefault(arg, getattr(self, arg))
		return type(self)(**kw)

	def layout(self, text, width, align, wrap):
		if align == self.ALIGN_INDENTATION:
			segs = self.calculate_text_segments_align(text, width, wrap)
			return self.align_layout(text, width, segs, wrap, self.ALIGN_LEFT)
		if wrap == self.WRAP_CUSTOM:
			segs = self.calculate_text_segments_align(text, width, wrap)
			return self.align_layout(text, width, segs, wrap, align)
		else:
			return super().layout(text, width, align, wrap)

	def calculate_text_segments_align(self, text, width, wrap):
		if wrap == self.WRAP_SPACE:
			wrap_after_characters = " "
			wrap_before_characters = " "
			wrap = self.WRAP_CUSTOM
		else:
			wrap_after_characters = self.wrap_after_characters
			wrap_before_characters = self.wrap_before_characters

		# ----- str vs bytes -----
		nl = "\n"
		align_character = self.align_character
		if urwid.compat.PYTHON3 and isinstance(text, bytes):
			nl = urwid.compat.B(nl)
			align_character = urwid.compat.B(align_character)
			wrap_after_characters = urwid.compat.B(wrap_after_characters)
			wrap_before_characters = urwid.compat.B(wrap_before_characters)
		wrap_after_characters += nl
		wrap_before_characters += nl

		# ----- align_character & indentation -----
		n = len(text)
		trailing_linebreaks = 0
		while text[n-1:n] == nl:
			trailing_linebreaks += 1
			n -= 1

		align_character_i0 = text.find(align_character)
		if align_character_i0 == -1:
			width_indent = 0
			align_character_i1 = -1
		else:
			align_character_i1 = align_character_i0 + len(align_character)
			i = text.rfind(nl, 0, align_character_i1)
			if i >= 0:
				i += 1
			else:
				i = 0
			width_indent = calc_width(text, i, align_character_i0)
			indent_i0 = i

			if self.auto_move_align_character:
				i1 = align_character_i1
				while text[i1:i1+1].isspace() and not text[i1:i1+1] == nl:
					i1 = move_next_char(text, i1, n)
				width_indent0 = width_indent
				width_indent1 = calc_width(text, align_character_i1, i1)
				width_indent = width_indent0 + width_indent1
				indent_i1 = i1
			else:
				width_indent0 = width_indent
				width_indent1 = 0
				indent_i1 = align_character_i1

			if width < width_indent + self.MIN_WIDTH:
				width_indent = 0

		# ----- process text -----
		layout_struct = []
		i0 = 0

		current_width = width
		current_indentation = 0
		while i0 < n:
			last_i0 = i0
			ln = []
			layout_struct.append(ln)

			# insert indentation
			if current_indentation:
				if self.repeat_prefix:
					if width_indent0:
						ln.append((width_indent0, indent_i0, align_character_i0))
					if width_indent1:
						ln.append((width_indent1, align_character_i1, indent_i1))
				elif self.EXPLICIT_SPACES:
					ln.append((current_indentation, i0, b" "*current_indentation))
				else:
					ln.append((current_indentation, None))

			# find end of line, by \n or lack of space
			i1 = text.find(nl, i0)
			if i1 == -1:
				i1 = n
			i1, sc = calc_text_pos(text, i0, i1, current_width)
			assert i1 >= i0

			# hide align_character
			if i0 <= align_character_i0 < i1:
				sc = calc_width(text, i0, align_character_i0)
				if sc:
					ln.append((sc, i0, align_character_i0))
				i0 = align_character_i1

				# move i1 right to make up for ignored align_character
				# I am starting at i0 to get the correct number of screen columns
				assert current_indentation == 0
				i = text.find(nl, i0)
				if i < 0:
					i = n
				i1, sc = calc_text_pos(text, i0, i, current_width-sc)

				current_indentation = width_indent
				current_width = width - current_indentation

			# apply wrap
			if i1 >= n or wrap == self.WRAP_CLIP:
				if sc:
					ln.append((sc, i0, i1))
				break
			elif wrap == self.WRAP_ELLIPSIS:
				if sc:
					ln.append((sc, i0, i1))
				self.add_ellipsis_character(text, ln, width)
				break
			elif text[i1:i1+1] == nl:
				if sc:
					ln.append((sc, i0, i1))
				i1 += 1
				i0 = i1
				continue
			elif wrap == self.WRAP_ANY:
				pass
			elif wrap == self.WRAP_CUSTOM:
				i = i1
				i1_min = i0 + 1
				while i > i1_min and text[i-1:i] not in wrap_after_characters and text[i:i+1] not in wrap_before_characters:
					i = move_prev_char(text, i1_min, i)
				if i > i1_min and not text[i0:i].isspace():
					i1 = i
				sc = calc_width(text, i0, i1)

			else:
				raise ValueError("invalid value for wrap: %s" % wrap)

			# add line
			if sc:
				ln.append((sc, i0, i1))

			while text[i1:i1+1].isspace():
				i1 = move_next_char(text, i1, n)

			# preprare processing of next line
			i0 = i1

			# stop if no progress has been made to avoid an infinite loop if window is too narrow
			if i0 == last_i0:
				break

		for i in range(trailing_linebreaks):
			layout_struct.append([])

		if not layout_struct:
			layout_struct.append([])

		return layout_struct

	def add_ellipsis_character(self, text, ln, width):
		if self.width_ellipsis_character > width:
			i0, sc = reverse_calc_text_pos(self.ellipsis_character, 0, len(self.ellipsis_character), width)
			ln.clear()
			ln.append((sc, 0, self.ellipsis_character[i0:]))
			return

		remaining_width = width - self.width_ellipsis_character
		ln_sc = sum(ts[0] for ts in ln)

		if ln_sc > remaining_width:
			assert ln
			while ln_sc > remaining_width:
				ts = ln[-1]
				del ln[-1]
				ln_sc -= ts[0]

			remaining_width -= ln_sc

			if len(ts) == 2:
				sc, i = ts
				ts = (remaining_width, i)
			elif isinstance(ts[2], int):
				sc, i0, i1 = ts
				i1, sc = calc_text_pos(text, i0, i1, remaining_width)
				ts = (sc, i0, i1)
			else:
				sc, i, insert_text = ts
				i1, sc = calc_text_pos(insert_text, 0, len(insert_text), remaining_width)
				ts = (sc, i, insert_text[:i1])

			if ts[0]:
				ln.append(ts)

		if not ln:
			i = 0
		else:
			i = ln[-1][1]

		if self.width_ellipsis_character:
			ln.append((self.width_ellipsis_character, i, self.ellipsis_character))

	def supports_align_mode(self, align):
		if align == self.ALIGN_INDENTATION:
			return True

		return super().supports_align_mode(align)

	def supports_wrap_mode(self, wrap):
		if wrap == self.WRAP_CUSTOM:
			return True

		return super().supports_wrap_mode(wrap)


WRAP = LogTextLayout.WRAP_SPACE
ALIGN = LogTextLayout.ALIGN_INDENTATION



#!/usr/bin/env python3

import math

def generate_keys(alphabet, num_keys):
	if num_keys == 0:
		return

	a,b, len_keys = calc_key_length(alphabet, num_keys)

	if len_keys > 1:
		keys = alphabet
		for i in range(len_keys - 2):
			keys = (key+c for key in keys for c in alphabet)
	else:
		assert a == 0
		keys = [""]
	
	
	keys = iter(keys)

	for i in range(a):
		yield next(keys)
	
	i = 0
	while True:
		key_start = next(keys)
		for c in alphabet:
			yield key_start + c
			i += 1
			if i >= b:
				return

def calc_key_length(alphabet, num_keys):
	len_alphabet = len(alphabet)
	len_key = math.ceil(math.log(num_keys, len_alphabet))
	a = math.floor((len_alphabet**len_key - num_keys) / (len_alphabet - 1))
	b = num_keys - a

	return a,b, len_key


if __name__ == '__main__':
	alph = "fg"
	n = 5
	#a,b, len_key = calc_key_length(alph, n)
	#print(f"{a}+{b}, {len_key}")
	out = list(generate_keys(alph, n))
	assert len(out) == n
	assert len(set(out)) == n
	for k in out:
		print(k)



#!/usr/bin/env python3

class History:

	"""
	There are two different functions to save data: commit and stash.
	For both is true:
	Given text will be saved if it is different from the current text.
	Given cursor position will always be saved.

	Committed text will never be changed.
	Committed cursor position can be changed.

	If the current entry was committed a new stash is inserted after it.
	If the current entry was stashed a new stash overwrites it.

	A commit is always appended to the end.
	All stashed changes are removed on the next commit.
	"""

	# data: list of entries
	# entry: 3-tuple (is_committed: bool, text: str, cursor_position: int)
	KEY_IS_COMMITTED = 0
	KEY_TEXT = 1
	KEY_CURSOR_POSITION = 2

	def __init__(self):
		self.data = [[False, "", 0]]
		self.index = 0

	# ---------- setter ----------

	def commit(self, text, cursor_position):
		"""remove all entries that are not committed and commit the given values"""
		self.data = [entry for entry in self.data if self.is_committed(entry)]
		self.index = self._get_last_index()

		if self.data and text == self.get_text():
			self.data[self.index][self.KEY_CURSOR_POSITION] = cursor_position
		else:
			self.index += 1
			self.data.append([True, text, cursor_position])

	def stash_and_prev(self, text, cursor_position):
		"""save the given values temporarily and select the previous entry"""
		self.stash(text, cursor_position)

		if self.index <= 0:
			return False
		else:
			self.index -= 1
			return True

	def stash_and_next(self, text, cursor_position):
		"""save the given values temporarily and select the next entry"""
		self.stash(text, cursor_position)

		if self.index >= self._get_last_index():
			return False
		else:
			self.index += 1
			return True

	# ---------- getter ----------

	def is_committed(self, entry=None):
		if entry is None:
			entry = self.data[self.index]

		return entry[self.KEY_IS_COMMITTED]

	def get_text(self, entry=None):
		if entry is None:
			entry = self.data[self.index]

		return entry[self.KEY_TEXT]

	def get_cursor_position(self, entry=None):
		if entry is None:
			entry = self.data[self.index]

		return entry[self.KEY_CURSOR_POSITION]

	def is_newest(self):
		return self.index == len(self.data) - 1

	def is_oldest(self):
		return self.index == 0

	# ---------- internal ----------

	def stash(self, text, cursor_position):
		if text == self.get_text():
			self.data[self.index][self.KEY_CURSOR_POSITION] = cursor_position
			return

		stash = [False, text, cursor_position]
		if self.is_committed():
			self.index += 1
			self.data.insert(self.index, stash)
		else:
			self.data[self.index] = stash

	def _get_last_index(self):
		return len(self.data) - 1

	# ---------- debug code ----------

	def format(self):
		table = []
		table.append(("", "committed?", "value", "cursor"))
		for i in range(len(self.data)):
			ln = []
			if i == self.index:
				ln.append("> ")
			else:
				ln.append("")

			ln.extend(str(v) for v in self.data[i])
			table.append(ln)

		col_widths = [len(max((table[row][col] for row in range(len(table))), key=len)) for col in range(len(table[0]))]

		out = []
		for row in table:
			ln = "| "
			for cell, width in zip(row, col_widths):
				ln += cell.ljust(width)
				ln += " | "
			out.append(ln)

		hline = "-" * (sum(col_widths) + (len(col_widths)-1)*3 + 4)

		out.insert(0, hline)
		out.insert(2, hline)
		out.append( hline)

		return "\n".join(out)



#!/usr/bin/env python3

import shlex
import argparse
import inspect
import io
import gettext
_ = gettext.gettext

from . import utils
from .api_subprocess import Runner


class ParseException(Exception):
	pass

class ArgumentParser(argparse.ArgumentParser):

	def error(self, message):
		raise ParseException(message)

class ErrorInCommand(Exception):
	pass


class CommandContainer:

	SEP = " "

	def __init__(self, app):
		self.commands = {}
		self.app = app

	def load_commands_from_module(self, module):
		self.load_commands_from_list(vars(module).values())

	def load_commands_from_list(self, classes):
		for var in classes:
			try:
				if not issubclass(var, Command):
					continue
			except TypeError:
				continue
			if var == Command:
				continue

			self.init_command_type(var)
			self.commands[var.get_name()] = var

	def init_command_type(self, cmd):
		cmd.app = self.app

		if cmd.run_before_init is None and cmd.run_after_init is None:
			cmd.run_before_init = False
			cmd.run_after_init = True
		elif cmd.run_before_init is None:
			cmd.run_before_init = False
		elif cmd.run_after_init is None:
			cmd.run_after_init = False

		cmd.create_parser()

	def split_cmd_args(self, command_line):
		assert isinstance(command_line, str)
		if self.SEP in command_line:
			command_name, args = command_line.split(self.SEP, 1)
		else:
			command_name = command_line
			args = ""
		return command_name, args

	def parse(self, command_line, silent=False):
		command_line = command_line.replace(r'\n', '\n')
		command_name, args = self.split_cmd_args(command_line)
		cmd = self.commands.get(command_name)
		if cmd is None and not silent:
			self.error(_("no such command: %s") % command_name)
		return cmd, args

	def execute(self, command_line, state=None):
		is_second_run = state == self.app.STATE_AFTER_INIT
		cmd, args = self.parse(command_line, silent=is_second_run)
		if cmd is None:
			return

		if state != self.app.STATE_RUNNING:
			if not cmd.run_before_init and not cmd.run_after_init:
				if not is_second_run:
					self.app.show_error(_("{cmd} cannot be used in config file").format(cmd=cmd.get_name()))
				return
			if state == self.app.STATE_BEFORE_INIT and not cmd.run_before_init:
				return
			if state == self.app.STATE_AFTER_INIT and not cmd.run_after_init:
				return

		try:
			cmd(args)
		except (ParseException, ErrorInCommand) as e:
			# error messages *from argparse* are internationalized
			# https://stackoverflow.com/a/35964548/2828064
			msg = self.app.pattern_error_in_command.format(err=e, cmd=command_line)
			self.app.show_error(msg, contains_align_character=True, contains_color_markup=True)

	def get_help(self, command_line):
		cmd, args = self.parse(command_line, silent=True)
		if cmd is None:
			mentioned_settings = []
			if self.is_primitive_command(command_line):
				out = self.get_help_primitive(command_line)
			else:
				out = [_("This command is undefined.")]
			return out, mentioned_settings

		out = cmd.get_help()
		out = out + cmd.get_config_file_behaviour()

		out, mentioned_settings = self.app.parse_help(out, parse_mentioned_settings=cmd.parse_mentioned_settings)
		return out, mentioned_settings

	def get_help_primitive(self, cmd):
		out = [_("This is a primitive command without further description.")]
		out.append("")
		out.append(_("possible usages:"))
		fmt = _("  {usage}")
		cmd = self.split_cmd_args(cmd)[0]
		for usage in sorted(self.app.primitive_commands):
			if self.split_cmd_args(usage)[0] == cmd:
				out.append(fmt.format(usage=usage))

		return out

	def is_primitive_command(self, cmd):
		cmd, args = self.split_cmd_args(cmd)
		for primitive_command in self.app.primitive_commands:
			primitive_command, args = self.split_cmd_args(primitive_command)
			if cmd == primitive_command:
				return True
		return False

	def error(self, msg):
		self.app.show_error(msg)

class Command(Runner):
	"""Abstract command class"""

	class AlignedLine:
		__slots__ = ('ln')

		def __init__(self, ln):
			self.ln = ln

	name = None
	run_before_init = None
	run_after_init = None
	parse_mentioned_settings = True

	# ------- class methods -------

	@classmethod
	def get_name(cls):
		classdict = cls.__mro__[0].__dict__
		if 'name' in classdict and classdict['name']:
			return cls.name
		return cls.__name__.replace("_", "-")

	@classmethod
	def create_parser(cls):
		cls.parser = ArgumentParser(prog=cls.get_name(), add_help=False, formatter_class=argparse.RawDescriptionHelpFormatter)
		cls.init_parser(cls.parser)

	@classmethod
	def init_parser(cls, parser):
		"""override this method if the command takes arguments"""
		pass

	@classmethod
	def init_help(cls, parser):
		"""override this method if you want to generate a help description"""
		doc = cls.__doc__
		if doc is None:
			doc = ""
		cls.parser.description = inspect.cleandoc(doc)

	@classmethod
	def init_help_lines(cls, parser):
		cls.init_help(parser)

		stdout = io.StringIO()
		cls.parser.print_help(stdout)
		cls.help_lines = stdout.getvalue().splitlines()

	@classmethod
	def get_help(cls):
		cls.init_help_lines(cls.parser)
		cls.init_help_lines = lambda parser: None
		return cls.help_lines

	@classmethod
	def get_config_file_behaviour(cls):
		out = []
		out.append("")
		out.append(_("behavior in config file:"))
		indent = "  "
		if cls.run_before_init and cls.run_after_init:
			out.append(indent + _("This command is executed before and after initialization."))
		elif cls.run_before_init:
			out.append(indent + _("This command is executed before initialization."))
		elif cls.run_after_init:
			out.append(indent + _("This command is executed after initialization."))
		else:
			out.append(indent + _("This command cannot be used in a config file."))
		return out


	# ------- instance methods -------

	def __init__(self, args):
		args = self.split_args(args)
		args = self.parse_args(args)
		self.execute(args)

	def split_args(self, args):
		try:
			return shlex.split(args)
		except ValueError as e:
			raise ParseException("Failed to split args of %s: %s" % (self.get_name(), e))

	def parse_args(self, args):
		return self.parser.parse_args(args)

	def execute(self, args):
		"""
		override this method

		args: argparse.Namespace instance
		"""
		pass


def quote(arg):
	return shlex.quote(arg)



#!/usr/bin/env python3

import shlex
import gettext
_ = gettext.gettext

from .urwid_text_layout import LogTextLayout as TextLayout

ATTR_SEP = '.'
GROUP_SEP = '.'

WC_HASH_ID = "{hash}"

BOOL_TRUE = "true"
BOOL_FALSE = "false"
VAL_AUTO = "auto"
VAL_APP = "app"

TYPE_COMMAND = "cmd"
TYPE_COLOR = "color"
TYPE_COLORED_STR = "coloredstr"

PATTERNS = {
	TYPE_COLOR : _("foreground[,emphases][/background]"),
}

VIEW_MODE_VER = "ver"
VIEW_MODE_HOR = "hor"
VIEW_MODE_ONE = "one"

BACKEND_CURSES = "curses"
BACKEND_RAW = "raw"

LOG_FOCUS_FIRST_LINE = "first-line"
LOG_FOCUS_UNTRACKED = "untracked-files"
LOG_FOCUS_UNSTAGED = "unstaged-changes"
LOG_FOCUS_STAGED = "staged-changes"
LOG_FOCUS_LATEST_COMMIT = "latest-commit"

COLOR_TITLE  = "color.title"
COLOR_SUBTITLE  = "color.subtitle"
COLOR_DECORATION_BRANCH_LOCAL = "color.title.commit.refnames.branch.local"
COLOR_DECORATION_BRANCH_REMOTE = "color.title.commit.refnames.branch.remote"
COLOR_DECORATION_TAG = "color.title.commit.refnames.tag"
COLOR_DECORATION_HEAD = "color.title.commit.refnames.head"
COLOR_DECORATION_HEAD_BRANCH_SEP = "color.title.commit.refnames.head-branch-sep"
COLOR_INFO = "color.echo.info"
COLOR_WARNING = "color.echo.warning"
COLOR_ERROR = "color.echo.error"
COLOR_SUCCESS = "color.echo.success"
COLOR_HINT = "color.link.hint"
COLOR_LINK = "color.link.hash"
COLOR_ELLIPSIS = "color.ellipsis"
COLOR_SPECIAL_CHARACTER = "color.special-character"
COLOR_KEY = "color.shortcut.key"
COLOR_KEY_PRESSED = "color.shortcut.key.pressed"
COLOR_MNEMONIC = "color.shortcut.mnemonic"
COLOR_CMD = "color.shortcut.cmd"
COLOR_KEY_IN_TEXT = "color.help.key"
COLOR_CMD_IN_TEXT = "color.help.cmd"
COLOR_CMD_IN_LINK = "color.link.cmd"
COLOR_LOG_STAGED = "color.log-entry.staged"
COLOR_LOG_UNSTAGED = "color.log-entry.unstaged"
COLOR_LOG_UNTRACKED = "color.log-entry.untracked"
COLOR_LOG_STASHES = "color.log-entry.stashes"
COLOR_DETAILS_UNTRACKED = "color.details-view.line-pattern-untracked"
COLOR_LINE_NUMBERS = "color.linenumber"
COLORS = (
	COLOR_TITLE, COLOR_SUBTITLE,
	COLOR_DECORATION_BRANCH_LOCAL, COLOR_DECORATION_BRANCH_REMOTE, COLOR_DECORATION_TAG, COLOR_DECORATION_HEAD,
	COLOR_DECORATION_HEAD_BRANCH_SEP,
	COLOR_INFO, COLOR_WARNING, COLOR_ERROR, COLOR_SUCCESS,
	COLOR_HINT, COLOR_LINK,
	COLOR_ELLIPSIS, COLOR_SPECIAL_CHARACTER,
	COLOR_KEY_PRESSED, COLOR_KEY, COLOR_CMD, COLOR_MNEMONIC,
	COLOR_CMD_IN_TEXT, COLOR_KEY_IN_TEXT, COLOR_CMD_IN_LINK,
	COLOR_LOG_STAGED, COLOR_LOG_UNSTAGED, COLOR_LOG_UNTRACKED, COLOR_LOG_STASHES,
	COLOR_DETAILS_UNTRACKED,
	COLOR_LINE_NUMBERS,
)

LOG_LEVEL_ALL = 0
LOG_LEVEL_INFO = 10
LOG_LEVEL_SUCCESS = 15
LOG_LEVEL_WARNING = 20
LOG_LEVEL_ERROR = 30
LOG_LEVEL_NONE = 40

LOG_LEVEL_NAME_ALL = "all"
LOG_LEVEL_NAME_INFO = "info"
LOG_LEVEL_NAME_SUCCESS = "success"
LOG_LEVEL_NAME_WARNING = "warning"
LOG_LEVEL_NAME_ERROR = "error"
LOG_LEVEL_NAME_NONE = "none"

LOG_LEVELS = (
	(LOG_LEVEL_NAME_ALL, LOG_LEVEL_ALL),
	(LOG_LEVEL_NAME_INFO, LOG_LEVEL_INFO),
	(LOG_LEVEL_NAME_SUCCESS, LOG_LEVEL_SUCCESS),
	(LOG_LEVEL_NAME_WARNING, LOG_LEVEL_WARNING),
	(LOG_LEVEL_NAME_ERROR, LOG_LEVEL_ERROR),
	(LOG_LEVEL_NAME_NONE, LOG_LEVEL_NONE),
)

RELATIVE_ROOT = 'root'
RELATIVE_GIT = 'git'
RELATIVE_CWD = 'cwd'

relative_values = (
	(RELATIVE_CWD, RELATIVE_CWD, _("paths relative to current working directory")),
	(RELATIVE_GIT, RELATIVE_GIT, _("paths relative to git top level directory")),
	(RELATIVE_ROOT, RELATIVE_ROOT, _("absolute paths")),
)

bool_values = ((BOOL_TRUE, True), (BOOL_FALSE, False))
align_values = (
	("left-indentation", TextLayout.ALIGN_INDENTATION, _("left align but add an indentation to broken lines for better readability. The amount of indentation is specified by the align character %app.align-character%. The align character is not displayed and is automatically removed or restored from the commands when changing this option. If this character occurs several times in a line only the first occurence is treated as align character and the others are displayed normally. If a line contains no align character it is assumed to be present before the first character. I have implemented this align mode myself. If it causes trouble try urwid's default left.")),
	("left", TextLayout.ALIGN_LEFT),
	("center", TextLayout.ALIGN_CENTER),
	("right", TextLayout.ALIGN_RIGHT),
)
wrap_values = (
	("any", TextLayout.WRAP_ANY, _("wrap lines, a line break is possible at any character")),
	("space", TextLayout.WRAP_SPACE, _("wrap lines, a line break is possible at spaces")),
	("clip", TextLayout.WRAP_CLIP, _("don't wrap lines")),
	("ellipsis", TextLayout.WRAP_ELLIPSIS, _("don't wrap lines, insert an ellipsis if a line cannot be displayed completely (requires urwid 2.1.0 or later)")),
	#TextLayout.WRAP_CUSTOM,
)

settings = {
	# key : (attribute, type/allowed values, help)
	# allowed values: a list of (valname, value, helpstr) tuples. helpstr can be omitted. A single str can be used instead of the tuple if value is a str equal to valname and there is no help.
	"layout.preferred" : ("app.preferred_view_mode", (VIEW_MODE_HOR, VIEW_MODE_VER, VIEW_MODE_ONE), "Specify which layout should be used (if the window is big enough)."),
	"layout.required-width-for-hor" : ("app.min_width_for_hor", int, "Do not switch to horizontal layout automatically if the window is smaller than the specified amount of columns. This does not apply when switching the layout manually."),
	"layout.required-height-for-ver" : ("app.min_height_for_ver", int, "Do not switch to vertical layout automatically if the window is smaller than the specified amount of rows. This does not apply when switching the layout manually."),
	"layout.auto-switch-on-resize" : ("app.auto_switch_view_mode_on_resize", bool, "Automatically switch to a more appropriate layout if the window size changes."),

	"log-view.show-untracked" : ("app.LogModel.show_untracked", bool, _("Show untracked files (if existing) in the log.")),
	"log-view.show-untracked-as-separate-group" : ("app.LogModel.untracked_files_as_separate_group", (
		(BOOL_TRUE,  True,  _("Show untracked files separate from unstaged changes.")),
		(BOOL_FALSE, False, _("Show untracked files together with unstaged changes.")),
	), _("This setting is ignored if %log-view.show-untracked% is disabled.")),
	"log-view.show-stashes" : ("app.LogModel.show_stashed_changes", bool, _("Show stashed changes (if existing) in the log.")),
	"log-view.show-stashes-as-group" : ("app.LogModel.stashed_changes_group", (
		(BOOL_TRUE,  True,  _("Group all stashes to one list item. Opening that list item opens a list of all stashes.")),
		(BOOL_FALSE, False, _("Show one list item for each stash.")),
		(VAL_AUTO,   None,  _("{true} if there is more than one stash, {false} otherwise").format(true=BOOL_TRUE, false=BOOL_FALSE)),
	), _("This setting is ignored if %log-view.show-stashes% is disabled.")),
	"log-view.show-stashes-in-reverse-order" : ("app.LogModel.stashed_changes_reversed_order", (
		(BOOL_TRUE,  True,  _("Show oldest stash on top.")),
		(BOOL_FALSE, False, _("Show newest stash on top.")),
	), _("This setting is ignored if %log-view.show-stashes% is disabled. You may want to change this setting if you change %log-view.show-stashes-as-group%.")),
	"log-view.default-focus" : ("app.LogView.default_focus_lines", (list, (LOG_FOCUS_UNTRACKED, LOG_FOCUS_UNSTAGED, LOG_FOCUS_STAGED, LOG_FOCUS_LATEST_COMMIT, LOG_FOCUS_FIRST_LINE)), _("This setting determines which line to select when the program starts. If the first element of this list is not contained in the log view try the next.")),
	"log-view.cmd.log" : ("app.LogModel.cmd_log", TYPE_COMMAND, "the command to generate the main part of the log view which shows the commits"),
	"log-view.cmd.stash-list" : ("app.LogModel.cmd_stash_list", TYPE_COMMAND, "the command to generate the list of stashes"),
	"log-view.cmd.unreachable" : ("app.LogModel.cmd_unreachable", TYPE_COMMAND, "the command to gather the unreachable commands if the command line argument --unreachable is used"),
	"log-view.line-pattern" : ("app.LogModel.ln_pattern", TYPE_COLORED_STR, _("how to format special log view entries such as staged and unstaged changes. If you change %log-view.cmd.log% you may want to adapt this. The value of this setting should include the wildcard {logentry}. Additionally to the color markup which can be used this is colored with one of %color.log-entry.*%.")),
	"log-view.line-pattern-error" : ("app.LogModel.ln_pattern_error", TYPE_COLORED_STR, _("this is displayed if one of the git commands %log-view.cmd.*% fails, the wild cards {err} and {cmd} are supported")),

	"details-view.auto-open" : ("app.is_auto_open_enabled", bool, "Automatically open the details when selecting a line. It may be useful to disable this if you are on a slow computer or are dealing with big commits."),
	"details-view.indent-broken-code" : ("app.DetailsModel.indent_broken_code", bool, "move the align character to the next non-whitespace character (this requires align=left-indentation and can be influenced by %app.tab%)"),
	"details-view.max-lines-per-file" : ("app.DetailsView.max_lines_per_file", int, "Trim long diffs to reduce loading time."),
	"details-view.max-lines-per-file.blob" : ("app.DetailsView.max_lines_per_file_blob", int, "Trim long blobs to reduce loading time."),
	"details-view.show-untracked-relative-to" : ("app.DetailsModel.untracked_relative", relative_values, _("how to show untracked files")),
	"details-view.cmd.commit" : ("app.DetailsModel.cmd_pattern", TYPE_COMMAND, _("the command which generates the details view for a commit, should include the wild card {wc_hash}").format(wc_hash=WC_HASH_ID)),
	"details-view.cmd.tag" : ("app.DetailsModel.cmd_pattern_tag", TYPE_COMMAND, _("the command which generates the details view for a tag, should include the wild card {wc_hash}").format(wc_hash=WC_HASH_ID)),
	"details-view.cmd.blob" : ("app.DetailsModel.cmd_pattern_blob", TYPE_COMMAND, _("the command which generates the details view for a blob, should include the wild card {wc_hash}").format(wc_hash=WC_HASH_ID)),
	"details-view.cmd.tree" : ("app.DetailsModel.cmd_pattern_tree", TYPE_COMMAND, _("the command which generates the details view for a tree, should include the wild card {wc_hash}").format(wc_hash=WC_HASH_ID)),
	"details-view.cmd.stash" : ("app.DetailsModel.cmd_pattern_stashed", TYPE_COMMAND, _("the command which generates the details view for a stash, should include the wild card {wc_hash}").format(wc_hash=WC_HASH_ID)),
	"details-view.cmd.staged" : ("app.DetailsModel.cmd_staged", TYPE_COMMAND, _("the command which generates the details view for staged changes")),
	"details-view.cmd.unstaged" : ("app.DetailsModel.cmd_unstaged", TYPE_COMMAND, _("the command which generates the details view for unstaged changes")),
	"details-view.cmd.untracked" : ("app.DetailsModel.cmd_untracked", TYPE_COMMAND, _("the command which generates the details view for untracked changes, each line is formatted with %details-view.unstaged.line-pattern-untracked% or %details-view.untracked.line-pattern-untracked% and colored with %color.details-view.line-pattern-untracked%")),
	"details-view.cmd.diff" : ("app.DiffModel.cmd", TYPE_COMMAND, _("the beginning of the command which generates the content for gitd")),
	"details-view.cmd.commit.header-data" : ("app.DetailsModel.cmd_start_header_data", TYPE_COMMAND, _("the beginning of the command which gathers the data for the function $commitheader, format and commit hash are appended automatically")),
	"details-view.cmd.commit.referenced-by" : ("app.DetailsModel.cmd_referencing_commits", TYPE_COMMAND, _("the beginning of the command which generates the list of commits referencing the current commit in the details view for a commit, --grep is added automatically, the output is formatted by %details-view.commit.line-pattern-referenced-by%")),
	"details-view.untracked.line-pattern-untracked" : ("app.DetailsModel.ln_pattern_untracked_in_separate_group", TYPE_COLORED_STR, _("how to format untracked files displayed in a separate group (if %log-view.show-untracked-as-separate-group% is enabled or there are no other unstaged changes), should include the wildcard {filename}, additionally to the color markup which can be used this is colored with %color.details-view.line-pattern-untracked%")),
	"details-view.unstaged.line-pattern-untracked" : ("app.DetailsModel.ln_pattern_untracked_in_unstaged_changes", TYPE_COLORED_STR, _("how to format untracked files displayed together with unstaged changes (if %log-view.show-untracked-as-separate-group% is disabled), should include the wildcard {filename}, additionally to the color markup which can be used this is colored with %color.details-view.line-pattern-untracked%")),
	"details-view.commit.line-pattern-referenced-by" : ("app.DetailsModel.ln_pattern_referencedby", str, _("how to format an entry in the list of commits which reference the current commit, should include the wildcard {logentry}, this is related to %log-view.line-pattern% but does not support color markup, this should not contain an %app.align-character% if %details-view.cmd.commit.referenced-by% contains one")),
	"details-view.commit.user-pattern" : ("app.DetailsModel.pattern_user", TYPE_COLORED_STR, _("how to display a user in $commitheader, supported are the wild cards {name} and {email}")),
	"details-view.commit.indentation-commit-message" : ("app.DetailsModel.indent_body", str, _("an indentation for the commit message in $commitheader")),
	"details-view.line-numbers.min-width" : ("app.DetailsModel.min_line_number_width", int, _("the amount of space to reserve for line numbers is determined for each hunk individually and depends on the biggest line number, this setting can be used to always reserve space for at least a certain number of digits so that the line numbers column has the same width for all hunks")),
	"details-view.line-numbers.sep" : ("app.DetailsModel.linenumber_suffix", TYPE_COLORED_STR, _("a separator between the line number and the line itself")),
	"details-view.line-numbers.sep-left-right" : ("app.DetailsModel.linenumber_sep", TYPE_COLORED_STR, _("a separator between the left and right line number (when showing the line numbers before a merge commit)")),

	#TODO: after changing these settings n and ? are not updated automatically, a new search must be started
	"search.case-sensitive" : ("app.search_case_sensitive", bool_values + ((VAL_AUTO, None, _("case sensitive if search text contains upper case characters, case insensitive otherwise")),), ""),
	"search.regex" : ("app.search_is_regex", bool, ""),

	"config.auto-reload" : ("app.is_auto_reload_config_enabled", bool, _("automatically load the config file after `config.edit`")),

	"app.display-module" : ("app.backend", (VAL_AUTO, BACKEND_RAW, BACKEND_CURSES), "http://urwid.org/manual/displaymodules.html#raw-and-curses-display-modules"),
	# does not work. switches focus without updating main_views_index or updating the details view. and it does not even handle scrolling. I don't like this anyway, so I'm not gonna spend time on this.
	#"urwid.handle-mouse" : ("app.handle_mouse", bool_values, _("{true} means that urwid handles mouse events. I hate this option because it means that you cannot select text with the mouse and copy it.").format(true=BOOL_TRUE, false=BOOL_FALSE)),
	"app.tab" : ("app.DetailsView.TAB", str, "urwid can't leave handling of tabs to the terminal, it needs to know exactly how wide a character is, otherwise it couldn't handle linebreaks correctly. This setting says what to display instead of a tab character."),
	"app.align" : ("app.align", align_values, _("how to align lines, see also %details-view.align%")),
	"details-view.align" : ("app.DetailsView.align", ((VAL_APP, VAL_APP, _("use same value like %app.align%")),)+align_values, _("how to align lines in details view")),
	"app.align-character" : ("app.align_character", str, _("If %app.align% is left-indentation this character specifies how far a broken line is supposed to be indented.")),
	"app.wrap" : ("app.wrap", wrap_values, _("whether or how to wrap lines, see also %details-view.wrap%")),
	"details-view.wrap" : ("app.DetailsView.wrap", ((VAL_APP, VAL_APP, _("use same value like %app.wrap%")),)+wrap_values, _("whether or how to wrap lines in details view")),
	"app.show-focus-in-all-views" : ("app.show_focus_in_all_views", (
		(BOOL_TRUE, True, _("the selected line is highlighted in all displayed views")),
		(BOOL_FALSE, False, _("the selected line is highlighted in the focused view only")),
	), _("only relevant if layout is {hor} or {ver}").format(hor=VIEW_MODE_HOR, ver=VIEW_MODE_VER)),
	"app.explicit-spaces" : ("app.explicit_spaces", bool, _("a workaround to avoid duplicates of full width characters appearing in the first column of indentation in some terminals with align=left-indentation. Requires urwid 2.1.0 or later. A side effect can be different highlighting of indentation in the selected row.")),
	"app.log-level" : ("app.log_level", LOG_LEVELS, _("show messages only if they have this or a higher importance (this refers to the status bar only, at the moment this program does not write any log files)")),
	"app.command-pipe-poll-time" : ("app.POLL_TIME_S", float, "Number of seconds. Time period in which it is checked whether something has been read from the command pipe."),

	"app.pattern.error-in-command" : ("app.pattern_error_in_command", TYPE_COLORED_STR, _("how to format an error message from a command, {err} is the error message and {cmd} is the command which was executed, this pattern is expected to contain an %app.align-character%, this is colored with %color.echo.error%")),
	"app.pattern.mnemonic" : ("app.pattern_mnemonic", TYPE_COLORED_STR, _("appended after a keyboard shortcut if it has a mnemonic, should include the wildcard {mnemonic}, this is colored with %color.shortcut.mnemonic%")),

	"link.hint-pattern" : ("app.Hintable.hint_pattern", TYPE_COLORED_STR, _("how to format the hint indicating which keys you need to press in order to follow a link, see `link`, this is colored using %color.link.hint%")),
	"link.hint-alphabet" : ("app.Hintable.HINTS_ALPHABET", str, _("the set of characters from which link hints are generated, earlier characters are preferred, must contain two characters minimum")),

	# I am not putting the titles in log-view or details-view because they are used in both
	"title.untracked" : ("app.model.title_untracked", str, _("log entry and details view title for untracked files (if %log-view.show-untracked% is true)")),
	"title.unstaged" : ("app.model.title_unstaged", str, _("log entry and details view title for unstaged changes")),
	"title.untracked-and-unstaged" : ("app.model.title_unstaged_and_untracked", str, _("log entry and details view title for a combination of untracked files and unstaged changes (if %log-view.show-untracked-as-separate-group% is false and %log-view.show-untracked% is true)")),
	"title.staged" : ("app.model.title_staged", str, _("log entry and details view title for staged changes")),
	"title.stashes" : ("app.model.title_stashes_group", str, _("log entry and details view title for the list of all stashes (if %log-view.show-stashes% and %log-view.show-stashes-as-group% are not false)")),
	# this is used in details-view only but seperating it from the other titles would not make sense
	"title.commit" : ("app.model.title_commit", str, _("title displayed in the details view for commits, the following wild cards are supported: {hash_id}, {refnames} (see %title.commit.refnames%), {subject}, {committer_name}, {committer_email}, {committer_date}, {author_name}, {author_email}, {author_date}, {body}")),
	"title.commit.refnames" : ("app.model.pattern_refnames", str, _("a pattern how to format {refnames} in %title.commit% if the commit has ref names (decorations), this pattern should include the wild card {refnames}")),
	"title.commit.refnames.sep" : ("app.model.refnames_sep", str, _("the seperator used in %title.commit.refnames% between different branches and tags")),
	"title.commit.refnames.head-branch-sep" : ("app.model.DetailsModel.decoration_head_branch_sep", str, _("the seperator used in %title.commit.refnames% between HEAD and the branch which HEAD is pointing to")),

	"help.pattern.cmd-with-key" : ("app.pattern_cmd_with_key", TYPE_COLORED_STR, _("how to format a command in the help if one or more keyboard shortcuts are mapped to the command, should contain the wildcards {cmd} or {keys}, cmd is colored with %color.help.cmd%, keys are colored with %color.help.key%")),
	"help.pattern.cmd-without-key" : ("app.pattern_cmd_without_key", TYPE_COLORED_STR, _("how to format a command in the help if no keyboard shortcut is mapped to the command, should contain the wildcard {cmd}, cmd is colored with %color.help.cmd%")),
	"help.key-sep" : ("app.sep_shortcut", str, _("the separator inserted between two keyboard shortcuts if several keyboard shortcuts are assigned to the same command")),
	"help.show-all-shortcuts" : ("app.help_show_all_shortcuts", (
		(BOOL_TRUE, True, _("show all keyboard shortcuts for the command regardless of the given arguments")),
		(BOOL_FALSE, False, _("show only those keyboard shortcuts which start with the same arguments")),
	), _("in the help of a command with arguments (when opened with `link` or `help --shortcuts`)")),
}

for color in COLORS:
	settings[color] = ("app.%s" % color.replace('.', '_'), TYPE_COLOR, "")



def keys():
	return sorted(settings.keys())

def get(key):
	attr, allowed_values, helpstr = settings[key]

	if allowed_values == bool:
		allowed_values = bool_values

	return attr, allowed_values, helpstr

def iter_allowed_values(allowed_values):
	for val in allowed_values:
		if isinstance(val, tuple):
			if len(val) == 3:
				valname, value, helpstr = val
			else:
				valname, value = val
				helpstr = None
		else:
			valname = val
			value = val
			helpstr = None
		yield valname, value, helpstr

def format_allowed_values(allowed_values):
	if isinstance(allowed_values, (tuple, list)):
		if allowed_values[0] == list:
			assert len(allowed_values) == 2
			return format_allowed_values(allowed_values[1])
		return ", ".join("%s" % valname for valname, value, helpstr in iter_allowed_values(allowed_values))

	elif allowed_values in PATTERNS:
		return PATTERNS[allowed_values]

	if not isinstance(allowed_values, str):
		allowed_values = allowed_values.__name__
	return "<%s>" % allowed_values

def label_allowed_values(allowed_values):
	if isinstance(allowed_values, (tuple, list)):
		if allowed_values[0] == list:
			assert len(allowed_values) == 2
			if isinstance(allowed_values[1], (tuple, list)):
				return _("A comma separated list which may contain the following values")
			else:
				return _("A comma separated list with data type")
		else:
			return _("Allowed values")
	elif allowed_values in PATTERNS:
		return _("Pattern")
	else:
		return _("Data type")

def format_value(value, allowed_values):
	if isinstance(allowed_values, (tuple, list)):
		if allowed_values[0] == list:
			return ",".join(format_value(v, allowed_values[1]) for v in value)

		for itervalname, itervalue, iterhelpstr in iter_allowed_values(allowed_values):
			if itervalue == value:
				return itervalname
		return repr(value)

	elif allowed_values == TYPE_COMMAND:
		if isinstance(value[0], str):
			value = [value]
		return '; '.join(shlex_join(cmd) for cmd in value)

	return str(value)

def can_be_true(allowed_values):
	if not isinstance(allowed_values, (tuple, list)):
		return False
	for valname, value, helpstr in iter_allowed_values(allowed_values):
		if valname == BOOL_TRUE:
			return True
	return False


def rgetattr(obj, name):
	if ATTR_SEP in name:
		n1, n2 = name.split(ATTR_SEP, 1)
		return rgetattr(getattr(obj, n1), n2)
	else:
		return getattr(obj, name)

def rsetattr(obj, name, value):
	if ATTR_SEP in name:
		n1, n2 = name.split(ATTR_SEP, 1)
		rsetattr(getattr(obj, n1), n2, value)
	elif hasattr(obj, name):
		setattr(obj, name, value)
	else:
		raise AttributeError("%s does not have attribute %s" % (obj, name))


def shlex_join(cmd):
	# this function is intended to be equivalent to shlex.join
	# I am not using shlex.join for backward compatibility with Python < 3.8
	# https://docs.python.org/3/library/shlex.html#shlex.join
	return " ".join(shlex.quote(w) for w in cmd)



#!/usr/bin/env python3

from git_viewer.api_commands import Command


# Despite this being a class I do not use camel case because
# the name of the class is used as the name of the command.
class my_command(Command):

	"""
	Each time a command is executed a new instance of this command is created.

	This doc string is displayed in the help of this command.
	Lines ending with a colon after an empty line are formatted with %color.subtitle%.

	If you reference other commands you should place them in `...` so that they are
	recognized as commands and highlighting, insertion of keyboard shortcuts and
	linking to their help is possible.

	If you reference settings you should place them in %...% so that they are recognized
	as settings and they are listed in the auto generated section "mentioned settings".
	"""

	# By default the name of the class is used
	# but if the name of the command shall contain
	# characters which are not allowed in a python
	# class name you can set this attribute.
	name = "my-command"

	# Specify if and when this command is allowed to be executed
	# in a config file at startup, see the help of `config.load`.
	# If run_before_init and run_after_init are not specified
	# run_before_init is initialized to False and run_after_init to True.
	# If only one of the two is specified the other one is initialized to False.
	run_before_init = None
	run_after_init = None


	# -------  methods which you will want to override -------

	@classmethod
	def init_parser(cls, parser):
		"""
		Define the arguments which this command can take.
		see the help of argparse.ArgumentParser
		"""
		pass

	def execute(self, args):
		"""
		Do what the command is supposed to do.
		args: argparse.Namespace instance

		This class inherits from Runner defined in api_subprocess.py
		which you can use to run sub processes.

		You have access to the main application object as self.app
		but no interface has really been defined yet.
		You can take a look at the code for App in main.py
		but I cannot guarantee that nothing will ever change.
		"""
		pass


	# -------  methods which you might want to override -------

	@classmethod
	def init_help(cls, parser):
		"""
		In this method you can generate help
		which is inserted directly after the description.

		This method is called once
		when the help for this command is opened for the first time.
		"""
		super().init_help(parser)

		parser.description += "\n"
		parser.description += "\n"
		parser.description += "numbers from one to ten:\n"
		parser.description += "  " + ", ".join("%d"%i for i in range(1,11))

	@classmethod
	def init_help_lines(cls, parser):
		"""
		In this method you can add another section after the arguments sections
		in which it is possible to align broken lines.

		This method is called once
		when the help for this command is opened for the first time.

		This method modifies a list of strings (or AlignedLine instances).
		Each element in this list is placed in a separate widget.
		Whether two lines are placed in one or two widgets makes a difference if visual mode is enabled.
		"""
		super().init_help_lines(parser)
		out = cls.help_lines

		out.append("")
		out.append("another section:")
		out.append("This section is inserted after the sections")
		out.append("for positional arguments and optional arguments")
		out.append("which are auto generated by argparse.")
		out.append("")
		out.append(cls.AlignedLine("- {align_character}In this section it is possible to use the align character to specify an indentation for broken lines.".format(align_character=cls.app.align_character)))

	@classmethod
	def get_help(cls):
		"""
		This method is called every time the help for this command is opened.
		In this method you could show the current value of settings.
		It returns a list, see init_help_lines.
		Do not modify the list returned by super().get_help()!
		Make sure the things you want to change are changed on a separate list object.
		"""
		return super().get_help()



#!/usr/bin/env python

import os
import stat
import tempfile
import re
import shlex
import shutil
import gettext
_ = gettext.gettext

from .api_subprocess import Runner, CommandError
from .string_formatter import formatter
from .color_decoder import ColorDecoder
from . import utils
from . import settings

ALIGN_CHAR = "^"

ID_UNTRACKED = "<untracked>"
ID_UNSTAGED = "<unstaged>"
ID_STAGED = "<staged>"
ID_STASHES_GROUP = "<stashed>"
SPECIAL_IDS = (ID_UNTRACKED, ID_UNSTAGED, ID_STAGED, ID_STASHES_GROUP)

TYPE_STASHED = "stashed"
TYPE_OTHER = "other"
TYPE_ERROR = "error"
TYPE_TAG = "tag"
TYPE_START_OF_FILE = "sof"
TYPE_NUMBERED_LINE = "numbered-line"
TYPE_UNTRACKED = "untracked"

#https://matthew-brett.github.io/curious-git/git_object_types.html
OBJECT_TYPE_COMMIT = "commit"
OBJECT_TYPE_TREE = "tree"
OBJECT_TYPE_BLOB = "blob"
OBJECT_TYPE_TAG = "tag"

title_untracked = _('New files, uncommitted and not checked in to index')
title_unstaged = _('Local uncommitted changes, not checked in to index')
title_unstaged_and_untracked = formatter.format(_('{title_unstaged} and {title_untracked:l}'), title_unstaged=title_unstaged, title_untracked=title_untracked)
title_staged   = _('Local changes checked in to index but not committed')
title_stashes_group = _('{number} stashes')
title_referencedby = _('This commit is referenced by the following commits:')
title_commit = _("commit {hash_id}{refnames}")
pattern_refnames = _(" ({refnames})")
refnames_sep = _(", ")

error_no_unreachable_commits = _('no unreachable commits')


# ---------- model classes ----------

class NameSpace:
	pass

class FileInfo:
	__slots__ = ('fn', 'object_id_before_left', 'object_id_before_right', 'object_id_after')

	def __init__(self, fn=None):
		self.fn = fn
		self.object_id_before_left = None
		self.object_id_before_right = None
		self.object_id_after = None

class LineInfo:
	__slots__ = ('linenumber_after', 'linenumber_before_right', 'linenumber_before_left',
		'formatted_linenumber_after', 'formatted_linenumber_before_right', 'formatted_linenumber_before_left',
		'is_modified')

	def __repr__(self):
		args = ", ".join("%s" % getattr(self, a) for a in ('linenumber_after', 'linenumber_before_right', 'linenumber_before_left'))
		return "%s(%s)" % (type(self).__name__, args)


class Model(Runner):

	"""
	A command is a list of strings as expected by the subprocess library
	or a list of such commands. If several commands are given they are
	executed one after another and their outputs are concatenated.
	_append_cmd, _remove_cmd and _toggle_cmd operate on the last command only.
	"""

	CARRIAGE_RETURN = '' #r'\r'

	cmd = None

	# commands starting with this prefix are not executed as shell commands
	# but looked up in command_functions
	INTERNAL_FUNCTION_PREFIX = "$"

	command_functions = dict()


	# ------- generate output -------

	def get_lines(self):
		if self.are_several_commands(self.cmd):
			out = list()
			for cmd in self.cmd:
				out.extend(self._get_lines(cmd))
		else:
			cmd = self.cmd
			out = self._get_lines(cmd)

		return out

	def _get_lines(self, cmd):
		i = len(self.INTERNAL_FUNCTION_PREFIX)
		if cmd[0][:i] == self.INTERNAL_FUNCTION_PREFIX:
			fun = cmd[0][i:]
			try:
				fun = self.command_functions[fun]
			except KeyError as e:
				supported_commands =  ", ".join(key for key in sorted(self.command_functions.keys()))
				raise CommandError(cmd, "no such internal command %r, should be one of %s" % (fun, supported_commands))
			args = cmd[1:]
			return fun(self, args)

		return self.run_and_get_output(cmd).replace('\r', self.CARRIAGE_RETURN).splitlines()


	# ------- manipulate commands -------

	@classmethod
	def replace_in_cmd(cls, cmd, wildcard, value):
		if not cls.are_several_commands(cmd):
			cmd = [cmd]

		return [[value if x == wildcard else x for x in subcmd] for subcmd in cmd]

	@classmethod
	def replace_characters_in_cmd(cls, cmd, old, new, count=-1):
		"""
		count = -1: replace all occurences of old with new
		count > -1: replace the first count occurences of old with new
		            in the first argument where old appears of every subcommand
		"""
		if cls.are_several_commands(cmd):
			return [cls.replace_characters_in_subcmd(subcmd, old, new, count) for subcmd in cmd]
		else:
			return cls.replace_characters_in_subcmd(cmd, old, new, count)

	@classmethod
	def replace_characters_in_subcmd(cls, subcmd, old, new, count=-1):
		out = []
		unfinished = True
		for arg in subcmd:
			if unfinished and old in arg:
				arg = arg.replace(old, new, count)
				if count >= 0:
					unfinished = False
			out.append(arg)
		return out


	@staticmethod
	def are_several_commands(cmd):
		return isinstance(cmd[0], list)

	@classmethod
	def _append_cmd(cls, cmd, options):
		if cls.are_several_commands(cmd):
			cmd[-1].extend(options)
		else:
			cmd.extend(options)

	@classmethod
	def _remove_cmd(cls, cmd, options):
		if cls.are_several_commands(cmd):
			cmd = cmd[-1]
		for opt in options:
			while opt in cmd:
				cmd.remove(opt)

	@classmethod
	def _toggle_cmd(cls, cmd, options):
		if cls.are_several_commands(cmd):
			cmd = cmd[-1]

		count_added = 0
		count_removed = 0
		for opt in options:
			if opt in cmd:
				while opt in cmd:
					cmd.remove(opt)
					count_removed += 1
			else:
				cmd.append(opt)
				count_added += 1

		return count_added, count_removed

	@classmethod
	def _copy_command(cls, cmd):
		if cls.are_several_commands(cmd):
			return [list(c) for c in cmd]
		return list(cmd)


	# ------- branches ------

	_local_branches=None
	def get_local_branches(self):
		cls = type(self)
		if cls._local_branches is None:
			cls._local_branches = self._get_branches()
		return cls._local_branches

	_remote_branches=None
	def get_remote_branches(self):
		cls = type(self)
		if cls._remote_branches is None:
			cls._remote_branches = self._get_branches("--remote")
		return cls._remote_branches

	def _get_branches(self, arg=None):
		cmd = ["git", "branch"]
		if arg:
			cmd.append(arg)

		stdout = self.run_and_get_output(cmd).splitlines()
		return [ln[2:] for ln in stdout]

	@classmethod
	def clear_cache(cls):
		cls._local_branches = None
		cls._remote_branches = None



class LogModel(Model):

	cmd_log = [["$status"], ["git", "log", "--color", "--graph", "--date=relative", "--pretty=format:%Cred%h%Creset -^%C(yellow)%d%Creset %s %Cgreen(%cd) %C(bold blue)<%an>%Creset", "--abbrev-commit"]]
	cmd_stash_list = ["git", "stash", "list", "--color", "--pretty=format:%C(cyan)* %Cred%gd%Creset - ^%s %Cgreen(%cr)%Creset", "--abbrev-commit"]
	cmd_unreachable = ["git", "fsck", "--unreachable"]

	@classmethod
	def append_cmd_log(cls, options):
		cmd = cls.cmd_log
		if cls.are_several_commands(cmd):
			cmd = cmd[-1]

		if "--no-walk" in options and "--graph" in cmd:
			cmd.remove("--graph")

		while "--no-graph" in options:
			# remove all occurences of --graph in cmd
			while "--graph" in cmd:
				cmd.remove("--graph")

			# remove all occurences of --graph in options before --no-graph
			i_no_graph = options.index("--no-graph")
			del options[i_no_graph]
			while True:
				try:
					i = options.index("--graph", 0, i_no_graph)
				except ValueError:
					break
				else:
					del options[i]
					i_no_graph -= 1

		cmd.extend(options)

	@classmethod
	def remove_cmd_log(cls, options):
		cls._remove_cmd(cls.cmd_log, options)

	@classmethod
	def toggle_cmd_log(cls, options):
		return cls._toggle_cmd(cls.cmd_log, options)


	def remove_align_character(self):
		align_char = ALIGN_CHAR
		cls = type(self)
		self.cmd_log = self.replace_characters_in_cmd(cls.cmd_log, align_char, "", 1)
		self.cmd_stash_list = self.replace_characters_in_cmd(cls.cmd_stash_list, align_char, "", 1)

		self.ln_pattern = cls.ln_pattern.replace(align_char, "", 1)
		self.ln_pattern_error = cls.ln_pattern_error.replace(align_char, "", 1)

	def restore_align_character(self):
		cls = type(self)
		self.cmd_log = cls.cmd_log
		self.cmd_stash_list = cls.cmd_stash_list

		self.ln_pattern = cls.ln_pattern
		self.ln_pattern_error = cls.ln_pattern_error

	WC_LOGENTRY = '{logentry}'
	ln_pattern = '* ^<color=default>' + WC_LOGENTRY
	ln_pattern_error = '\n<color=%color.echo.error%>{err}</color>\nsee %{setting}%\ncmd = {cmd}\n'

	show_untracked = True
	untracked_files_as_separate_group = False

	show_stashed_changes = True
	stashed_changes_reversed_order = False
	stashed_changes_group = None  # True|False|None. None = auto

	re_hex = r'[0-9A-F]'
	reo_hash_id = re.compile(r"(?:\b|(?<=m))(%(re_hex)s{7,40})\b" % dict(re_hex=re_hex), re.IGNORECASE)
	reo_stash_id = re.compile(r"stash@\{\d+\}", re.IGNORECASE)

	def get_lines(self):
		"""return a list of 3-tuples (line, id, type), each representing one line"""
		# committed changes
		self.cmd = self.cmd_log
		try:
			lines = super().get_lines()
		except CommandError as e:
			log = self.format_command_error(e, setting='log-view.cmd.log')
		else:
			log = lines
			log = [(ln, self.extract_hash_id(ln), TYPE_OTHER) if isinstance(ln, str) else ln for ln in log]

		return log

	def format_command_error(self, e, setting):
		ln = self.ln_pattern_error.format(err=e.err, cmd=e.cmd, setting=setting)
		ln = utils.colored_str_to_markup(ln, self.app.define_color)
		return [(ln, None, TYPE_ERROR)]

	def status(self, args):
		# uncommitted changes
		if args:
			cmd = ['$status'] + args
			raise CommandError(cmd, "status takes no arguments")

		self.cmd = ["git", "status", "--porcelain=v1"]
		status = super().get_lines()

		untracked = False
		unstaged = False
		staged = False

		# X: index
		# Y: work tree
		for ln in status:
			if ln[0] not in ' !?':
				staged = True
			if ln[1] not in ' !?':
				unstaged = True
			if ln[:2] == '??':
				untracked = True

		if not self.show_untracked:
			untracked = False

		log = []
		if staged:
			ln = self.format_log_entry(settings.COLOR_LOG_STAGED, title_staged)
			log.insert(0, (ln, ID_STAGED, TYPE_OTHER))

		if unstaged:
			if untracked and not self.untracked_files_as_separate_group:
				title = title_unstaged_and_untracked
				untracked = False
			else:
				title = title_unstaged
			ln = self.format_log_entry(settings.COLOR_LOG_UNSTAGED, title)
			log.insert(0, (ln, ID_UNSTAGED, TYPE_OTHER))

		if untracked:
			ln = self.format_log_entry(settings.COLOR_LOG_UNTRACKED, title_untracked)
			log.insert(0, (ln, ID_UNTRACKED, TYPE_OTHER))


		# stashed changes
		if self.show_stashed_changes:
			self.cmd = self.cmd_stash_list
			try:
				stashed = super().get_lines()
			except CommandError as e:
				log = self.format_command_error(e, setting='log-view.cmd.stash-list') + log
			else:
				if not stashed:
					pass
				elif self.stashed_changes_group or (self.stashed_changes_group is None and len(stashed) > 1):
					ln = formatter.format(title_stashes_group, number=len(stashed))
					ln = self.format_log_entry(settings.COLOR_LOG_STASHES, ln)
					log.insert(0, (ln, ID_STASHES_GROUP, TYPE_STASHED))
				else:
					log = self.format_stashed(stashed) + log

		return log

	status.arguments = ''
	status.description = _('show stashed, untracked, unstaged and staged changes if existing')
	status.settings = [
		'log-view.show-untracked',
		'log-view.show-untracked-as-separate-group',
		'log-view.cmd.stash-list',
		'log-view.show-stashes',
		'log-view.show-stashes-as-group',
		'log-view.show-stashes-in-reverse-order',
		'log-view.line-pattern',
		'title.stashes',
		'title.untracked',
		'title.unstaged',
		'title.untracked-and-unstaged',
		'title.staged',
		settings.COLOR_LOG_STASHES,
		settings.COLOR_LOG_UNTRACKED,
		settings.COLOR_LOG_UNSTAGED,
		settings.COLOR_LOG_STAGED,
	]

	command_functions = dict(
		status = status,
	)

	def format_log_entry(self, color, logentry):
		ln = self.ln_pattern.format(logentry=logentry)
		ln = utils.colored_str_to_markup(ln, self.app.define_color)
		return (color, ln)

	def format_stashed(self, stashed):
		if self.stashed_changes_reversed_order:
			stashed = reversed(stashed)

		log = list()
		for ln in stashed:
			stash_id = self.extract_stash_id(ln)
			log.append((ln, stash_id, TYPE_STASHED))

		return log

	@classmethod
	def extract_hash_id(cls, text):
		m = cls.reo_hash_id.search(text)
		if m:
			return m.group(0)
		else:
			return None

	@classmethod
	def extract_stash_id(cls, text):
		m = cls.reo_stash_id.search(text)
		if m:
			return m.group(0)
		else:
			return None

class StashesModel(LogModel):

	def get_lines(self):
		self.cmd = self.cmd_stash_list
		try:
			stashed = Model.get_lines(self)
		except CommandError as e:
			return self.format_command_error(e, setting='log-view.cmd.stash-list')

		return self.format_stashed(stashed)

class UnreachableModel(LogModel):

	@classmethod
	def append_cmd_unreachable(cls, options):
		cls._append_cmd(cls.cmd_unreachable, options)

	def get_lines(self):
		self.cmd = self.cmd_unreachable
		try:
			lines = Model.get_lines(self)
		except CommandError as e:
			ln = self.ln_pattern_error.format(err=e.err, cmd=e.cmd)
			ln = utils.colored_str_to_markup(ln, self.app.define_color)
			return [(ln, None, TYPE_ERROR)]

		lines = [ln.split() for ln in lines]
		lines = [ln[2] for ln in lines if ln[0] == "unreachable" and ln[1] == "commit"]
		if not lines:
			err = error_no_unreachable_commits + "\n"
			ln = self.ln_pattern_error.format(err=err, cmd=self.cmd)
			ln = utils.colored_str_to_markup(ln, self.app.define_color)
			return [(ln, None, TYPE_ERROR)]

		if self.are_several_commands(LogModel.cmd_log):
			self.cmd_log = self._copy_command(LogModel.cmd_log[-1])
		else:
			self.cmd_log = self._copy_command(LogModel.cmd_log)
		self.append_cmd_log.__func__(self, ["--no-walk"])
		self.cmd_log += lines
		log = LogModel.get_lines(self)
		return log


class DetailsModel(Model):

	WC_HASH_ID = settings.WC_HASH_ID

	cmd_pattern = [
		#["git", "show", "--color", "--decorate", "--no-patch", WC_HASH_ID],
		["$commitheader", WC_HASH_ID],
		["$referencedby", WC_HASH_ID],
		["git", "show", "--color", "--decorate", "--format=format:", "--patch", WC_HASH_ID],
	]
	cmd_start_header_data = ["git", "show"]
	cmd_pattern_tag = ["git", "show", "--color", "--format=%ncommit %H%n%cd", "--no-patch", WC_HASH_ID, "--"]
	cmd_pattern_blob = ["git", "show", WC_HASH_ID]
	cmd_pattern_tree = ["git", "show", "--color", WC_HASH_ID]
	cmd_unstaged = ["git", "diff", "--color"]
	cmd_staged = cmd_unstaged + ["--cached"]
	cmd_untracked = ["$untracked"]
	cmd_pattern_stashed = [
			["git", "log", "-g", "-1", "--pretty=format:%C(yellow)%gD%Creset%ncommit %H%nAuthor: %an <%ae>%nDate:   %ad%w(,4,4)%n%+B%n%n", "--color", WC_HASH_ID],
			["git", "stash", "show", "--patch", "--color", WC_HASH_ID],
	]
	cmd_referencing_commits = [w for w in LogModel.cmd_log[-1] if w != "--graph"] + ["--branches"]

	log_format_options = (
		'--no-decorate',
		'--decorate',
		'--decorate-refs',
		'--decorate-refs-exclude',
		'--use-mailmap',

		'--date-order',
		'--author-date-order',
		'--topo-order',
		'--reverse',

		'--pretty',
		'--format',
		'--abbrev-commit',
		'--no-abbrev-commit',
		'--oneline',
		'--encoding',

		'--expand-tabs',
		'--no-expand-tabs',

		'--notes',
		'--no-notes',

		'--show-signature',

		'--relative-date',
		'--date',

		'--left-right',
	)

	_diff_commands = [cmd_pattern, cmd_unstaged, cmd_staged, cmd_pattern_stashed, cmd_start_header_data]


	@classmethod
	def append_cmd_diff(cls, options):
		for cmd in cls._diff_commands:
			cls._append_cmd(cmd, options)

	@classmethod
	def remove_cmd_diff(cls, options):
		for cmd in cls._diff_commands:
			cls._remove_cmd(cmd, options)

	@classmethod
	def toggle_cmd_diff(cls, options):
		sum_added = 0
		sum_removed = 0
		for cmd in cls._diff_commands:
			added, removed = cls._toggle_cmd(cmd, options)
			sum_added += added
			sum_removed += removed

		return sum_added, sum_removed


	@classmethod
	def append_cmd_log(cls, options):
		for opt in options:
			# values are given in the same argument, separated by a =
			key = opt.split('=',1)[0]
			if key in cls.log_format_options:
				cls.cmd_referencing_commits.append(opt)

	@classmethod
	def remove_cmd_log(cls, options):
		cls._remove_cmd(cls.cmd_referencing_commits, options)

	@classmethod
	def toggle_cmd_log(cls, options):
		return cls._toggle_cmd(cls.cmd_referencing_commits, options)


	def remove_align_character(self):
		self.insert_align_character = False

	def restore_align_character(self):
		self.insert_align_character = True


	decoration_head_branch_sep = " -> "

	ln_pattern_referencedby = "* {logentry}"
	ln_pattern_untracked_in_unstaged_changes = _("untracked:<color=default> {filename}</color>")
	ln_pattern_untracked_in_separate_group = _("{filename}")

	pattern_user = _("{name} <{email}>")

	linenumber_new = "{new}"
	linenumber_old = "{old}"
	linenumber_old_right = "{oldright}"
	linenumber_old_left = "{oldleft}"
	linenumber_sep = " "
	linenumber_off = ""
	linenumber_suffix = " "
	linenumber = linenumber_new
	min_line_number_width = 3

	re_ansi_esscape_sequence = "(\[[0-9;]*m)?"
	reo_start_new_file = re.compile(re_ansi_esscape_sequence + "diff (--git a/(?P<fn1>.*?) b/|--cc (?P<fn2>[^]*)|--combined (?P<fn3>[^]*))")
	reo_object_id = re.compile(re_ansi_esscape_sequence + "index ((?P<beforeleft>[0-9a-f]+),)?(?P<before>[0-9a-f]+)\.\.(?P<after>[0-9a-f]+)")
	reo_linenumber = re.compile(re_ansi_esscape_sequence +
		r"(@@@ -(?P<start_before_left>\d+)(,(?P<number_before_left>\d+))? -(?P<start_before_right>\d+)(,(?P<number_before_right>\d+))?"
		r"|@@ -(?P<start_before>\d+)(,(?P<number_before>\d+))?) "
		r"\+(?P<start_after>\d+)(,(?P<number_after>\d+))? @@"
	)
	reo_normal_line_exists_after = re.compile(re_ansi_esscape_sequence + r"[+ ]")
	reo_normal_line_existed_before = re.compile(re_ansi_esscape_sequence + r"[- ]")
	reo_merge_line_exists_after = re.compile(re_ansi_esscape_sequence + r"[+ ][+ ]")
	reo_merge_line_existed_left = re.compile(re_ansi_esscape_sequence  + r"([- ][+ ]|--)")
	reo_merge_line_existed_right = re.compile(re_ansi_esscape_sequence + r"([+ ][- ]|--)")

	reo_hash_id = LogModel.reo_hash_id
	reo_url = re.compile(r'\S+?([.]\S+)+')

	untracked_relative = settings.RELATIVE_CWD

	# if linenumbers are activated: insert ALIGN_CHAR in every line which is not represented by a tuple
	insert_align_character = True

	# if insert_align_character is activated: auto_move_align_character for lines in diff (not commit message)
	indent_broken_code = False

	def __init__(self, hash_id, id_type):
		self.hash_id = hash_id
		self.id_type = id_type
		self.is_merge = None
		self.is_diff = True

	def set_cmd(self):
		self.setting_cmd = None
		if self.hash_id == ID_STAGED:
			self.cmd = self.cmd_staged
			self.setting_cmd = 'details-view.cmd.staged'
			self.title = [(settings.COLOR_TITLE, title_staged)]
		elif self.hash_id == ID_UNSTAGED:
			self.cmd = self.cmd_unstaged
			self.setting_cmd = 'details-view.cmd.unstaged'
			self.title = [(settings.COLOR_TITLE, title_unstaged)]
			self.show_untracked_in_unstaged = LogModel.show_untracked and not LogModel.untracked_files_as_separate_group
		elif self.hash_id == ID_UNTRACKED:
			self.cmd = self.cmd_untracked
			self.setting_cmd = 'details-view.cmd.untracked'
			self.title = [(settings.COLOR_TITLE, title_untracked)]
		elif self.id_type == TYPE_STASHED:
			self.cmd = self.replace_in_cmd(self.cmd_pattern_stashed, self.WC_HASH_ID, self.hash_id)
			self.setting_cmd = 'details-view.cmd.stash'
			self.title = None
		else:
			try:
				self.object_type = get_object_type(self.hash_id)
			except CommandError as e:
				self.out.extend(self.format_command_error(e, _("get the object type")))
				self.object_type = None
			if self.object_type == OBJECT_TYPE_TAG:
				self.cmd = self.replace_in_cmd(self.cmd_pattern_tag, self.WC_HASH_ID, self.hash_id)
				self.setting_cmd = 'details-view.cmd.tag'
				self.title = None
				self.is_diff = False
			elif self.object_type == OBJECT_TYPE_TREE:
				self.cmd = self.replace_in_cmd(self.cmd_pattern_tree, self.WC_HASH_ID, self.hash_id)
				self.setting_cmd = 'details-view.cmd.tree'
				self.title = None
				self.is_diff = False
			elif self.object_type == OBJECT_TYPE_BLOB:
				self.cmd = self.replace_in_cmd(self.cmd_pattern_blob, self.WC_HASH_ID, self.hash_id)
				self.setting_cmd = 'details-view.cmd.blob'
				self.title = None
				self.is_diff = False
			else: # OBJECT_TYPE_COMMIT
				self.cmd = self.replace_in_cmd(self.cmd_pattern, self.WC_HASH_ID, self.hash_id)
				self.setting_cmd = 'details-view.cmd.commit'
				self.title = None

	def get_lines(self):
		"""
		returns a list of lines where each line is one of the following:
		- "text"
		- [("attr", "text")]  # urwid markup, must not contain tabs!
		- ("text", type)
		- ("text", type, auto_move_align_character)
		- ("text", TYPE_UNTRACKED, file_name)
		- ("text", TYPE_START_OF_FILE, file_info)
		- ("text", TYPE_NUMBERED_LINE, auto_move_align_character, line_info)
		"""
		self.object_type = None
		self.out = []
		self.set_cmd()
		out = self.out

		try:
			out.extend(super().get_lines())
		except CommandError as e:
			out.extend(self.format_command_error(e, setting=self.setting_cmd))
			return out

		if self.hash_id == ID_UNSTAGED and self.show_untracked_in_unstaged:
			cmd = self.cmd
			self.cmd = self.cmd_untracked
			try:
				untracked = super().get_lines()
			except CommandError as e:
				untracked = self.format_command_error(e, setting='details-view.cmd.untracked')
			else:
				pattern = self.ln_pattern_untracked_in_unstaged_changes
				untracked = self.format_untracked(pattern, untracked)
			finally:
				self.cmd = cmd

			if untracked:
				untracked.append("")
				out = untracked + out
				self.title = [(settings.COLOR_TITLE, title_unstaged_and_untracked)]
		elif self.hash_id == ID_UNTRACKED:
			pattern = self.ln_pattern_untracked_in_separate_group
			out = self.format_untracked(pattern, out)

		if self.is_diff:
			self.insert_linenumbers(out)

		if self.title:
			out.insert(0, self.title)
			out.insert(1, "")

		return out

	def format_untracked(self, pattern, filenames):
		path = lambda x: x
		if self.cmd_untracked == ['$untracked']:
			if self.untracked_relative == settings.RELATIVE_CWD:
				path = os.path.abspath
		return [([(settings.COLOR_DETAILS_UNTRACKED, utils.colored_str_to_markup(formatter.format(pattern, filename=fn), self.app.define_color))], TYPE_UNTRACKED, FileInfo(path(fn))) for fn in filenames]

	def format_command_error(self, e, action=None, setting=None):
		cmd = "cmd = %r\n" % (e.cmd,)
		err = e.err
		if action:
			err += "\n" + _("while trying to {do_action}").format(do_action=action)
		if setting:
			err += "\n" + _("see %{setting}%").format(setting=setting)

		return [cmd, (err, TYPE_ERROR)]

	def insert_linenumbers(self, out):
		self.file_info = None
		ns = NameSpace()
		ns.out = out
		ns.look_for_hint = True
		for ns.i in range(len(ns.out)):
			ln = ns.out[ns.i]
			if isinstance(ln, tuple):
				ns.i += 1
			elif isinstance(ln, list):
				ns.i += 1
			elif ns.look_for_hint:
				self.insert_linenumbers__process_metainfo_line(ns)
			else:
				self.insert_linenumbers__process_content_line(ns)

	def insert_linenumbers__process_metainfo_line(self, ns):
		ns.line_info = None
		ln = ns.out[ns.i]
		m = self.reo_linenumber.match(ln)
		if m:
			if m.group('start_before') is not None:
				self.is_merge = False
				self._set_linenumber(ns, 'before_right', m, 'before')
				ns.linenumber_before_left = ns.linenumber_before_right
				ns.last_linenumber_before_left = ns.last_linenumber_before_right
				ns.fmt_linenumber_before_left = ns.fmt_linenumber_before_right

				ns.reo_line_exists_after = self.reo_normal_line_exists_after
				ns.reo_line_existed_left = self.reo_normal_line_existed_before
				ns.reo_line_existed_right = self.reo_normal_line_existed_before
				ns.i_align_character = 1
			else:
				self.is_merge = True
				self._set_linenumber(ns, 'before_right', m, 'before_right')
				self._set_linenumber(ns, 'before_left', m, 'before_left')

				ns.reo_line_exists_after = self.reo_merge_line_exists_after
				ns.reo_line_existed_left = self.reo_merge_line_existed_left
				ns.reo_line_existed_right = self.reo_merge_line_existed_right
				ns.i_align_character = 2

			self._set_linenumber(ns, 'after', m, 'after')

			ns.look_for_hint = False
		else:
			m = self.reo_object_id.match(ln)
			if m:
				self.file_info.object_id_before_left = m.group('beforeleft')
				self.file_info.object_id_before_right = m.group('before')
				self.file_info.object_id_after = m.group('after')


		if self.insert_align_character:
			ns.out[ns.i] = ALIGN_CHAR + ln
		if self.is_start_new_file(ln):
			ns.out[ns.i] = (ns.out[ns.i], TYPE_START_OF_FILE, self.file_info)

	def _set_linenumber(self, ns, ns_name, match, re_name):
		linenumber = match.group('start_'+re_name)
		linenumber = int(linenumber)
		number_lines = match.group('number_'+re_name)
		if number_lines is None:
			last_linenumber = linenumber
		else:
			last_linenumber = linenumber + int(number_lines) - 1
		fmt = self.create_number_format(last_linenumber)
		setattr(ns, 'linenumber_'+ns_name, linenumber)
		setattr(ns, 'last_linenumber_'+ns_name, last_linenumber)
		setattr(ns, 'fmt_linenumber_'+ns_name, fmt)

	def create_number_format(self, last_linenumber):
		width = len(str(last_linenumber))
		width = max(width, self.min_line_number_width)
		return "%" + str(width) + "s"

	def insert_linenumbers__process_content_line(self, ns):
		orig_ln = ns.out[ns.i]
		ln = orig_ln

		if self.insert_align_character:
			ln = self.insert_align_char(ln, ns.i_align_character)

		ns.line_info = LineInfo()
		ln = (ln, TYPE_NUMBERED_LINE, self.indent_broken_code, ns.line_info)
		ns.out[ns.i] = ln

		self.increment_line_number_and_set_line_info(orig_ln, ns)

	def increment_line_number_and_set_line_info(self, orig_ln, ns):
		ns.line_info.linenumber_after = ns.linenumber_after
		ns.line_info.linenumber_before_right = ns.linenumber_before_right
		ns.line_info.linenumber_before_left = ns.linenumber_before_left
		ns.line_info.is_modified = False
		no_increment = True

		if ns.reo_line_exists_after.match(orig_ln):
			ns.line_info.formatted_linenumber_after = ns.fmt_linenumber_after % ns.linenumber_after
			ns.linenumber_after += 1
			no_increment = False
		else:
			ns.line_info.formatted_linenumber_after = ns.fmt_linenumber_after % ""
			ns.line_info.is_modified = True

		if ns.reo_line_existed_left.match(orig_ln):
			ns.line_info.formatted_linenumber_before_left = ns.fmt_linenumber_before_left % ns.linenumber_before_left
			ns.linenumber_before_left += 1
			no_increment = False
		else:
			ns.line_info.formatted_linenumber_before_left = ns.fmt_linenumber_before_left % ""
			ns.line_info.is_modified = True

		if ns.reo_line_existed_right.match(orig_ln):
			ns.line_info.formatted_linenumber_before_right = ns.fmt_linenumber_before_right % ns.linenumber_before_right
			ns.linenumber_before_right += 1
			no_increment = False
		else:
			ns.line_info.formatted_linenumber_before_right = ns.fmt_linenumber_before_right % ""
			ns.line_info.is_modified = True

		if no_increment:
			# The usual way of counting lines does not work.
			# This happens with --word-diff.
			# So I'm assuming each line is present in all versions.
			# Of course this is not always fullfilled.
			# But it's still better than giving each line the same number.
			ns.line_info.formatted_linenumber_after = ""
			ns.line_info.formatted_linenumber_before_left = ""
			ns.line_info.formatted_linenumber_before_right = ""
			ns.linenumber_after += 1
			ns.linenumber_before_left += 1
			ns.linenumber_before_right += 1

		if ns.linenumber_after > ns.last_linenumber_after and ns.linenumber_before_right > ns.last_linenumber_before_right and ns.linenumber_before_left > ns.last_linenumber_before_left:
			ns.look_for_hint = True

	@staticmethod
	def insert_align_char(ln, pos_align_character):
		n_escape = 0
		for m in ColorDecoder.reo_color_code.finditer(ln):
			i = m.start()
			pos = i - n_escape
			if pos >= pos_align_character:
				break
			n_escape += m.end() - i

		i_align_character = pos_align_character + n_escape
		return ln[:i_align_character] + ALIGN_CHAR + ln[i_align_character:]

	def is_start_new_file(self, ln):
		"""ln: str, possibly starting with one ansi escape sequence"""
		if isinstance(ln, list):
			return False
		m = self.reo_start_new_file.match(ln)
		if m:
			fn = m.group('fn1')
			if fn is None:
				fn = m.group('fn2')
				if fn is None:
					fn = m.group('fn3')
			self.file_info = FileInfo()
			self.file_info.fn = fn
			return True
		else:
			return False


	# ------- command functions ------

	def referencedby(self, args):
		assert len(args) == 1
		hash_id = args[0]

		cmd = ["git", "show", "--format=%h", "--no-patch", hash_id]
		short_hash = self.run_and_get_output(cmd).strip()

		bak = self.cmd
		self.cmd = self.cmd_referencing_commits +  ["--grep", r"\<"+short_hash]
		try:
			out = super().get_lines()
		except CommandError as e:
			msg = "command for searching referencing commits failed:\ncmd = {e.cmd}\n{e.err}\nreturn code: {e.returncode}".format(e=e)
			return [(msg, TYPE_ERROR)]
		self.cmd = bak

		out = [(self.ln_pattern_referencedby.format(logentry=ln), TYPE_OTHER) for ln in out]
		if out:
			out.insert(0, title_referencedby)
			out.insert(0, "")

		return out

	referencedby.arguments = 'CURRENT_COMMIT_HASH'
	referencedby.description = _('show a list of commits referencing the current commit')
	referencedby.settings = ['details-view.cmd.commit.referenced-by', 'details-view.commit.line-pattern-referenced-by']

	def untracked(self, args):
		assert len(args) == 0

		self.cmd = ["git", "status", "--porcelain=v1"]
		status = super().get_lines()
		root = get_git_root()

		files = []

		# X: index
		# Y: work tree
		for ln in status:
			if ln[:2] == '??':
				fn = ln[3:]
				if self.untracked_relative == settings.RELATIVE_ROOT or self.untracked_relative == settings.RELATIVE_CWD:
					fn = os.path.join(root, fn)
					if self.untracked_relative == settings.RELATIVE_CWD:
						fn = os.path.relpath(fn)
				files.append(fn)

		return files

	untracked.arguments = ''
	untracked.description = _('return untracked files')
	untracked.settings = ['details-view.show-untracked-relative-to']

	indent_body = _("    ")
	def commitheader(self, args):
		assert len(args) == 1
		hash_id = args[0]

		out = []
		values = self.get_commit_header_data(hash_id)
		if values is None:
			return []
		refnames = values.pop('refnames')
		if refnames:
			markup_refnames = utils.replace_to_markup(refnames, ', ', refnames_sep, markup_function=self.markup_deco)
			markup_refnames = utils.replace_to_markup(pattern_refnames, '{refnames}', markup_refnames)
		else:
			markup_refnames = ""

		title = title_commit.format(refnames='{refnames}', **values)
		markup_title = utils.replace_to_markup(title, '{refnames}', markup_refnames)
		markup_title = [(settings.COLOR_TITLE, markup_title)]
		out.append(markup_title)
		self.append_author_committer(out, values)
		out.append("")
		out.append(self.indent_body + values["subject"])
		body = values["body"].rstrip()
		if body:
			out.append("")
			out.append(self.indent_body + body)
		return out

	def get_commit_header_data(self, hash_id):
		key_fmt = ">>> %s:"
		keys = (
			("hash_id", "%H"),
			("refnames", "%D"),
			("subject", "%s"),
			("committer_name", "%cn",),
			("committer_email", "%ce",),
			("committer_date", "%cd",),
			("author_name", "%an",),
			("author_email", "%ae",),
			("author_date", "%ad",),
			("body", "%b"),
			("EOF", "") # ignore trailing newline while ensuring that body is added to values
		)
		fmt = "\n".join((key_fmt % key) + "\n" + wc for key,wc in keys)

		self.cmd = self.cmd_start_header_data + ["--no-patch", "--format="+fmt, hash_id]
		try:
			lines = super().get_lines()
		except CommandError as e:
			self.out.extend(self.format_command_error(e, setting='details-view.cmd.commit.header-data'))
			return None

		values = {}
		current_key = None
		next_key_index = 0
		for ln in lines:
			if next_key_index < len(keys) and ln == key_fmt % keys[next_key_index][0]:
				current_key = keys[next_key_index][0]
				next_key_index += 1
			else:
				if current_key not in values:
					values[current_key] = ln
				else:
					values[current_key] += "\n" + ln

		return values

	def append_author_committer(self, out, values):
		fmt = self.pattern_user
		author = fmt.format(name=values['author_name'], email=values['author_email'])
		committer = fmt.format(name=values['committer_name'], email=values['committer_email'])
		author_date = values['author_date']
		committer_date = values['committer_date']

		align_char = ALIGN_CHAR if self.insert_align_character else ""

		t = []
		if author == committer:
			t.append((_("Author/Committer"), align_char + author))
		else:
			t.append((_("Author"), align_char + author))

		if author_date == committer_date:
			t.append((_("Date"), align_char + author_date))
		else:
			t.append((_("Author Date"), align_char + author_date))

		if committer != author:
			t.append((_("Committer"), align_char + committer))

		if committer_date != author_date:
			t.append((_("Committer Date"), align_char + committer_date))

		for ln in utils.format_table(t, fmt=['%s:']):
			ln = utils.colored_str_to_markup(ln, define_color=self.app.define_color)
			# wrap line in a tuple so that ALIGN_CHAR is not added again
			ln = (ln, None)
			out.append(ln)

	def markup_deco(self, decoration):
		HEAD_SEP = " -> "
		i = decoration.find(HEAD_SEP)
		if i >= 0:
			head = decoration[:i]
			sep = self.decoration_head_branch_sep
			i += len(HEAD_SEP)
			branch = decoration[i:]
			return [(settings.COLOR_DECORATION_HEAD, head), (settings.COLOR_DECORATION_HEAD_BRANCH_SEP, sep), (settings.COLOR_DECORATION_BRANCH_LOCAL, branch)]

		if decoration.startswith("tag: "):
			return (settings.COLOR_DECORATION_TAG, decoration)

		if decoration in self.get_local_branches():
			return (settings.COLOR_DECORATION_BRANCH_LOCAL, decoration)

		if decoration in self.get_remote_branches():
			return (settings.COLOR_DECORATION_BRANCH_REMOTE, decoration)

		return decoration

	commitheader.arguments = 'CURRENT_COMMIT_HASH'
	commitheader.description = _('show the meta info of the current commit (hash, branch, tag, author, committer, date, message)')
	commitheader.settings = [
		'details-view.cmd.commit.header-data',
		'details-view.commit.user-pattern',
		'title.commit.refnames.head-branch-sep',
		settings.COLOR_TITLE,
		settings.COLOR_DECORATION_HEAD,
		settings.COLOR_DECORATION_HEAD_BRANCH_SEP,
		settings.COLOR_DECORATION_BRANCH_LOCAL,
		settings.COLOR_DECORATION_BRANCH_REMOTE,
		settings.COLOR_DECORATION_TAG,
	]


	command_functions = dict(
		referencedby = referencedby,
		untracked = untracked,
		commitheader = commitheader,
	)


class DiffModel(DetailsModel):

	cmd = ['git', 'diff', '--color']

	def __init__(self, args):
		self.hash_id = None
		self.id_type = None
		self.title = None

		self.args = args
		self.is_diff = True

		self.is_merge = None

	def set_cmd(self):
		self.setting_cmd = "details-view.cmd.diff"
		self.cmd = type(self).cmd + self.args

DetailsModel._diff_commands.append(DiffModel.cmd)


# ---------- modifiers ----------

def append_cmd_log(log_args):
	LogModel.append_cmd_log(log_args)
	DetailsModel.append_cmd_log(log_args)

def remove_cmd_log(log_args):
	LogModel.remove_cmd_log(log_args)
	DetailsModel.remove_cmd_log(log_args)

def toggle_cmd_log(log_args):
	sum_added = 0
	sum_removed = 0
	for cls in (LogModel, DetailsModel):
		added, removed = cls.toggle_cmd_log(log_args)
		sum_added += added
		sum_removed += removed

	return sum_added, sum_removed


def append_cmd_unreachable(unreachable_args):
	UnreachableModel.append_cmd_unreachable(unreachable_args)


def append_cmd_diff(diff_args):
	DetailsModel.append_cmd_diff(diff_args)

def remove_cmd_diff(diff_args):
	DetailsModel.remove_cmd_diff(diff_args)

def toggle_cmd_diff(log_args):
	return DetailsModel.toggle_cmd_diff(log_args)


# ---------- getters ----------

def get_containing_tag(hash_id, exclusive):
	cmd = ["git", "tag", "--list", "--contains", hash_id, "--sort", "version:refname"]
	out = Runner().run_and_get_output(cmd)
	out = out.rstrip()
	out = out.splitlines()
	if exclusive:
		for tag in out:
			if is_same_commit(tag, hash_id):
				continue
			return tag
		return None

	if out:
		return out[0]
	else:
		return None

def get_last_tag(hash_id, exclusive):
	if exclusive:
		hash_id = "%s^" % hash_id
	cmd = ["git", "describe", "--abbrev=0", hash_id]
	try:
		out = Runner().run_and_get_output(cmd)
	except CommandError:
		return None

	out = out.rstrip()
	return out

def is_same_commit(rev1, rev2):
	return get_commit_hash(rev1) == get_commit_hash(rev2)

def get_commit_hash(revision):
	# show does not work because it shows the annotation for tags
	cmd = ["git", "log", "-1", "--pretty=format:%H", revision, "--"]
	out = Runner().run_and_get_output(cmd)
	out = out.rstrip()
	return out

def get_object_type(hash_id):
	cmd = ["git", "cat-file", "-t", hash_id]
	out = Runner().run_and_get_output(cmd)
	out = out.rstrip()
	return out

def get_git_version():
	cmd = ["git", "--version"]
	try:
		out = Runner().run_and_get_output(cmd)
	except CommandError:
		return None

	out = out.rstrip()
	return out

def get_git_root():
	cmd = ["git", "rev-parse", "--show-toplevel"]
	try:
		out = Runner().run_and_get_output(cmd)
	except CommandError:
		return None

	out = out.rstrip()
	return out

def get_relative_path(path):
	'''
	path: a path relative to the git root directory
	returns: a path relative to the current working directory
	'''
	if os.path.isabs(path):
		return path
	root = get_git_root()
	cwd = os.getcwd()
	cwd = os.path.relpath(cwd, root) + os.sep
	return os.path.relpath(path, cwd)


# ---------- opener ----------

def initattr(obj, name, value):
	if hasattr(obj, name):
		return
	setattr(obj, name, value)

class Opener:

	def __init__(self):
		settings
		self._map = {}
		self.add_editor("vi", linenumber="+{ln}", readonly="-R")
		self.add_editor("vim", linenumber="+{ln}", readonly="-R")
		self.add_editor("nano", linenumber="+{ln}", readonly="-v")

	def add_editor(self, name, command=None, *, linenumber="", readonly=""):
		if not command:
			command = name
		self._map[command] = name
		setattr(self,  "editor_%s_command" % name, command)
		initattr(self, "editor_%s_linenumber" % name, linenumber)
		initattr(self, "editor_%s_readonly" % name, readonly)
		settings.settings["editor.%s.command" % name] = ("app.model.opener.editor_%s_command" % name, str, _("the value of the EDITOR environment variable which activates the corresponding editor settings"))
		settings.settings["editor.%s.linenumber" % name] = ("app.model.opener.editor_%s_linenumber" % name, str, _("a command line argument which is added if a specific line shall be selected, the wild card {ln} specifies the line number"))
		settings.settings["editor.%s.readonly" % name] = ("app.model.opener.editor_%s_readonly" % name, str, _("a command line argument which is added if a file shall be opened read only"))

	def get_editor(self):
		cmd = os.environ.get('EDITOR', None)
		if cmd:
			name = self._map.get(cmd, None)
			if name is None:
				# in case EDITOR is an absolute path
				name = self._map.get(os.path.split(cmd)[1], None)
		else:
			name = "vi"
			cmd = self.get_editor_argument(name, "command")
			assert cmd
		return name, cmd

	def get_editor_argument(self, editor_name, argument_name):
		return getattr(self, "editor_%s_%s" % (editor_name, argument_name), "")

	def open_old_version_in_editor(self, fn, commit, object_id, which, line_number):
		fn = os.path.split(fn)[1]
		basename,ext = os.path.splitext(fn)
		tmpfn = "{fn}@{which}-{commit}{ext}".format(fn=basename, which=which, commit=commit, ext=ext)
		runner = Runner()
		with tempfile.TemporaryDirectory() as tmpdir:
			tmpfn = os.path.join(tmpdir, tmpfn)
			with open(tmpfn, 'wt') as f:
				cmd = ['git', 'show', object_id]
				content = runner.run_and_get_output(cmd)
				f.write(content)
			os.chmod(tmpfn, stat.S_IREAD)
			self.open_file_in_editor(f.name, line_number, read_only=True)

	def open_file_in_editor(self, fn, line_number, read_only=False, create_dirs=False):
		if not os.path.exists(fn):
			if create_dirs:
				path = os.path.split(os.path.abspath(fn))[0]
				os.makedirs(path, exist_ok=True)
			else:
				raise FileNotFoundError(fn)
		if line_number == 0:
			line_number = 1
		runner = Runner()
		editor_name, cmd = self.get_editor()
		if line_number is not None:
			cmd += " " + self.get_editor_argument(editor_name, "linenumber").format(ln=line_number)
		if read_only:
			cmd += " " + self.get_editor_argument(editor_name, "readonly")
		cmd += " " + shlex.quote(fn)
		runner.run_interactive(cmd, shell=True)

	def check_editor(self):
		editor_name, cmd = self.get_editor()
		exe = shlex.split(cmd)[0]
		which = shutil.which(exe)
		if not which:
			if editor_name is None:
				editor_name = _("<unknown editor>")
			return _("cannot open editor {editor_name}, no such executable {exe!r}").format(exe=exe, editor_name=editor_name)
	
		return None


opener = Opener()
open_old_version_in_editor = opener.open_old_version_in_editor
open_file_in_editor = opener.open_file_in_editor
check_editor = opener.check_editor


if __name__ == '__main__':
	log = LogModel()
	for ln in log.get_lines():
		print(ln)



#!/usr/bin/env python3

import urwid

# see Urwid Documentation, Release 1.3.1
# section 3.8.2 Foreground and Background Settings


class COLOR:

	# ------- foreground -------

	FG_DEFAULT       = 'default'

	FG_BLACK         = 'black'
	FG_RED           = 'dark red'
	FG_GREEN         = 'dark green'
	FG_YELLOW        = 'brown'
	FG_BLUE          = 'dark blue'
	FG_MAGENTA       = 'dark magenta'
	FG_CYAN          = 'dark cyan'
	FG_WHITE         = 'light gray'

	FG_BRIGHT_BLACK   = 'dark gray'
	FG_BRIGHT_RED     = 'light red'
	FG_BRIGHT_GREEN   = 'light green'
	FG_BRIGHT_YELLOW  = 'yellow'
	FG_BRIGHT_BLUE    = 'light blue'
	FG_BRIGHT_MAGENTA = 'light magenta'
	FG_BRIGHT_CYAN    = 'light cyan'
	FG_BRIGHT_WHITE   = 'white'


	# ------- background -------

	BG_DEFAULT = 'default'

	BG_BLACK   = 'black'
	BG_RED     = 'dark red'
	BG_GREEN   = 'dark green'
	BG_YELLOW  = 'brown'
	BG_BLUE    = 'dark blue'
	BG_MAGENTA = 'dark magenta'
	BG_CYAN    = 'dark cyan'
	BG_WHITE   = 'light gray'


class EMPH:

	"""usage: COLOR.FG_* + EMPH.BOLD"""

	BOLD      = ",bold"
	UNDERLINE = ",underline"
	STANDOUT  = ",standout"



if __name__ == '__main__':

	# ------- test -------

	fg_colors = [
		COLOR.FG_BLACK,
		COLOR.FG_RED,
		COLOR.FG_GREEN,
		COLOR.FG_YELLOW,
		COLOR.FG_BLUE,
		COLOR.FG_MAGENTA,
		COLOR.FG_CYAN,
		COLOR.FG_WHITE,
	]

	fg_colors_light = [
		COLOR.FG_LIGHT_BLACK,
		COLOR.FG_LIGHT_RED,
		COLOR.FG_LIGHT_GREEN,
		COLOR.FG_LIGHT_YELLOW,
		COLOR.FG_LIGHT_BLUE,
		COLOR.FG_LIGHT_MAGENTA,
		COLOR.FG_LIGHT_CYAN,
		COLOR.FG_LIGHT_WHITE,
	]

	bg_colors = [
		COLOR.BG_BLACK,
		COLOR.BG_RED,
		COLOR.BG_GREEN,
		COLOR.BG_YELLOW,
		COLOR.BG_BLUE,
		COLOR.BG_MAGENTA,
		COLOR.BG_CYAN,
		COLOR.BG_WHITE,
	]

	def input_handler(key):
		if key == "q" or key == "enter":
			raise urwid.ExitMainLoop()

		return key

	indent = "    "
	exp_text_1 = "hello "
	exp_text_2 = "world"

	widgets = [urwid.Text("foreground colors:")] \
	        + [urwid.Text([indent, ("fg: %s" % col, exp_text_1), ("fg: %s" % col_light, exp_text_2)]) for col, col_light in zip(fg_colors, fg_colors_light)] \
	        + [urwid.Text("")] \
	        + [urwid.Text("background colors:")] \
	        + [urwid.Text([indent, ("bg: %s" % col, exp_text_1 + exp_text_2)]) for col in bg_colors]

	palette = [("fg: %s" % col, col, COLOR.BG_DEFAULT) for col in fg_colors] \
	        + [("fg: %s" % col, col, COLOR.BG_DEFAULT) for col in fg_colors_light] \
	        + [("bg: %s" % col, COLOR.FG_DEFAULT, col) for col in bg_colors]

	listbox = urwid.ListBox(urwid.SimpleFocusListWalker(widgets))
	urwid.MainLoop(listbox, palette=palette, unhandled_input=input_handler).run()



#!/usr/bin/env python3

"""
It does not seem to be possible to pass through colors in urwid or curses.
Therefore ColorDecoder.decode translates a string containing ansi color
escape sequences into a list of tuples (<colorname>, <text>).
(Even though it sounds a little like the raw_display might support color
 pass through, it does not. In fact, it is the default screen.)

Please note that the returned list can be empty
because tuples are dropped if <text> is empty.
Passing an empty list to an urwid Text widget
causes an Exception.

Implemented escape sequences are:
- foreground colors (3[0-7])
- background colors (4[0-7])
- bright foreground (9[0-7])
- reset (empty)
- bold (1)
https://stackoverflow.com/a/33206814
"""

import re

class ColorDecoder():

	# ---------- input ---------

	# matches:
	#     [m       (reset)
	#     [1m      (bold)
	#     [31m     (red)
	#     [1;31m   (light red)
	# invalid:
	#     [131m
	#
	# so in general I want to match
	# `a`, `b` and `a;b` but not `ab`.
	# I cannot use `a|b|a;b` because
	# `a` and `b` contain named groups.
	# Instead I'm adding a lookahead to
	# `a` (series) that the next character
	# must not be a number.
	reo_color_code = re.compile(
		r'\['
		r'((?P<series>[01])(?![0-9]))?'
		r';?'
		r'((?P<fgbg>[349])(?P<color>[0-79]))?'
		r'm'
	)

	NORMAL = '0'
	BOLD   = '1'

	FOREGROUND = '3'
	BACKGROUND = '4'
	FOREGROUND_LIGHT = '9'

	COLOR_DEFAULT = '9'
	COLOR_BLACK   = '0'
	COLOR_RED     = '1'
	COLOR_GREEN   = '2'
	COLOR_YELLOW  = '3'
	COLOR_BLUE    = '4'
	COLOR_MAGENTA = '5'
	COLOR_CYAN    = '6'
	COLOR_GRAY    = '7'


	# ---------- output ---------

	FG_DEFAULT       = 'default'

	FG_BLACK         = 'black'
	FG_RED           = 'red'
	FG_GREEN         = 'green'
	FG_YELLOW        = 'yellow'
	FG_BLUE          = 'blue'
	FG_MAGENTA       = 'magenta'
	FG_CYAN          = 'cyan'
	FG_GRAY          = 'gray'

	FG_LIGHT_BLACK   = 'dark gray'
	FG_LIGHT_RED     = 'light red'
	FG_LIGHT_GREEN   = 'light green'
	FG_LIGHT_YELLOW  = 'light yellow'
	FG_LIGHT_BLUE    = 'light blue'
	FG_LIGHT_MAGENTA = 'light magenta'
	FG_LIGHT_CYAN    = 'light cyan'
	FG_LIGHT_GRAY    = 'white'


	BG_DEFAULT = 'default'

	BG_BLACK   = 'black'
	BG_RED     = 'red'
	BG_GREEN   = 'green'
	BG_YELLOW  = 'yellow'
	BG_BLUE    = 'blue'
	BG_MAGENTA = 'magenta'
	BG_CYAN    = 'cyan'
	BG_GRAY    = 'gray'


	# ---------- mapping ---------

	decode_map = {
		FOREGROUND : {
			COLOR_DEFAULT : FG_DEFAULT,
			COLOR_BLACK   : FG_BLACK,
			COLOR_RED     : FG_RED,
			COLOR_GREEN   : FG_GREEN,
			COLOR_YELLOW  : FG_YELLOW,
			COLOR_BLUE    : FG_BLUE,
			COLOR_MAGENTA : FG_MAGENTA,
			COLOR_CYAN    : FG_CYAN,
			COLOR_GRAY    : FG_GRAY,
		},
		FOREGROUND_LIGHT : {
			COLOR_BLACK   : FG_LIGHT_BLACK,
			COLOR_RED     : FG_LIGHT_RED,
			COLOR_GREEN   : FG_LIGHT_GREEN,
			COLOR_YELLOW  : FG_LIGHT_YELLOW,
			COLOR_BLUE    : FG_LIGHT_BLUE,
			COLOR_MAGENTA : FG_LIGHT_MAGENTA,
			COLOR_CYAN    : FG_LIGHT_CYAN,
			COLOR_GRAY    : FG_LIGHT_GRAY,
		},
		BACKGROUND : {
			COLOR_DEFAULT : BG_DEFAULT,
			COLOR_BLACK   : BG_BLACK,
			COLOR_RED     : BG_RED,
			COLOR_GREEN   : BG_GREEN,
			COLOR_YELLOW  : BG_YELLOW,
			COLOR_BLUE    : BG_BLUE,
			COLOR_MAGENTA : BG_MAGENTA,
			COLOR_CYAN    : BG_CYAN,
			COLOR_GRAY    : BG_GRAY,
		},
	}

	@staticmethod
	def combine_colors(series, fg, bg):
		return "%s-%s:%s" % (series, fg, bg)


	# ---------- instance methods ---------

	def __init__(self):
		self.reset()

	def reset(self):
		self.color_fg = self.COLOR_DEFAULT
		self.color_bg = self.COLOR_DEFAULT
		self.series = self.NORMAL
		self.foreground = self.FOREGROUND
		self.background = self.BACKGROUND
	
	def decode(self, text):
		out = []

		i = None
		color = self.get_current_color()
		for m in self.reo_color_code.finditer(text):
			self.apply_color(m)

			subtext = text[i:m.start()]
			if subtext:
				out.append((color, subtext))

			color = self.get_current_color()
			i = m.end()

		subtext = text[i:]
		if subtext:
			out.append((color, subtext))

		return out

	def apply_color(self, color_code_match):
		m = color_code_match

		fgbg = m.group("fgbg")
		color = m.group("color")
		series = m.group("series")

		if fgbg is None and series is None:
			self.reset()

		if fgbg == self.FOREGROUND or fgbg == self.FOREGROUND_LIGHT:
			self.color_fg = color
			self.foreground = fgbg
		elif fgbg == self.BACKGROUND:
			self.color_bg = color
			self.background = fgbg

		if series is not None:
			self.series = series
	
	def get_current_color(self):
		fg = self.decode_map[self.foreground][self.color_fg]
		bg = self.decode_map[self.background][self.color_bg]
		return self.combine_colors(self.series, fg, bg)


if __name__ == '__main__':
	import sys

	decoder = ColorDecoder()

	if len(sys.argv) <= 1 or sys.argv[1] == '--full':
		for ln in sys.stdin.readlines():
			l = decoder.decode(ln)
			print(l)

	else:
		colors = set()
		for ln in sys.stdin.readlines():
			l = decoder.decode(ln)
			for color, text in l:
				colors.add(color)

		print("\n".join(sorted(colors)))
